from __future__ import annotations

from typing import Dict, List, Optional, Any
from deltacat.utils.metrics import MetricsConfig
from deltacat.utils.common import ReadKwargsProvider
from deltacat.io.object_store import IObjectStore
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.compute.compactor import DeltaAnnotated


class HashBucketInput(Dict):
    @staticmethod
    def of(
        annotated_delta: DeltaAnnotated,
        primary_keys: List[str],
        num_hash_buckets: int,
        num_hash_groups: int,
        enable_profiler: Optional[bool] = False,
        metrics_config: Optional[MetricsConfig] = None,
        read_kwargs_provider: Optional[ReadKwargsProvider] = None,
        object_store: Optional[IObjectStore] = None,
        deltacat_storage=unimplemented_deltacat_storage,
        deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    ) -> HashBucketInput:

        result = HashBucketInput()
        result["annotated_delta"] = annotated_delta
        result["primary_keys"] = primary_keys
        result["num_hash_buckets"] = num_hash_buckets
        result["num_hash_groups"] = num_hash_groups
        result["enable_profiler"] = enable_profiler
        result["metrics_config"] = metrics_config
        result["read_kwargs_provider"] = read_kwargs_provider
        result["object_store"] = object_store
        result["deltacat_storage"] = deltacat_storage
        result["deltacat_storage_kwargs"] = deltacat_storage_kwargs or {}

        return result

    @property
    def annotated_delta(self) -> DeltaAnnotated:
        return self["annotated_delta"]

    @property
    def primary_keys(self) -> List[str]:
        return self["primary_keys"]

    @property
    def num_hash_buckets(self) -> int:
        return self["num_hash_buckets"]

    @property
    def num_hash_groups(self) -> int:
        return self["num_hash_groups"]

    @property
    def enable_profiler(self) -> Optional[bool]:
        return self.get("enable_profiler")

    @property
    def metrics_config(self) -> Optional[MetricsConfig]:
        return self.get("metrics_config")

    @property
    def read_kwargs_provider(self) -> Optional[ReadKwargsProvider]:
        return self.get("read_kwargs_provider")

    @property
    def object_store(self) -> Optional[IObjectStore]:
        return self.get("object_store")

    @property
    def deltacat_storage(self) -> unimplemented_deltacat_storage:
        return self.get("deltacat_storage")

    @property
    def deltacat_storage_kwargs(self) -> Optional[Dict[str, Any]]:
        return self.get("deltacat_storage_kwargs")



from __future__ import annotations

from typing import Dict, Any, Optional, List
from pyarrow.parquet import FileMetaData


class PartialFileDownloadParams(Dict[str, Any]):
    """
    A content type params class used to represent arguments required
    to down the file partially. This is useful specifically in cases
    where you'd like to instruct downloader to partially download a
    manifest entry.
    """

    pass


class PartialParquetParameters(PartialFileDownloadParams):
    @staticmethod
    def of(
        row_groups_to_download: Optional[List[int]] = None,
        num_row_groups: Optional[int] = None,
        num_rows: Optional[int] = None,
        in_memory_size_bytes: Optional[float] = None,
        pq_metadata: Optional[FileMetaData] = None,
    ) -> PartialParquetParameters:

        if (
            row_groups_to_download is None
            or num_row_groups is None
            or num_rows is None
            or in_memory_size_bytes is None
        ):
            assert (
                pq_metadata is not None
            ), "Parquet file metadata must be passed explicitly"

            num_row_groups = pq_metadata.num_row_groups
            row_groups_to_download = [rg for rg in range(num_row_groups)]
            in_memory_size_bytes = 0.0
            num_rows = pq_metadata.num_rows

            for rg in row_groups_to_download:
                row_group_meta = pq_metadata.row_group(rg)
                in_memory_size_bytes += row_group_meta.total_byte_size

        result = PartialParquetParameters(
            {
                "row_groups_to_download": row_groups_to_download,
                "num_row_groups": num_row_groups,
                "num_rows": num_rows,
                "in_memory_size_bytes": in_memory_size_bytes,
            }
        )

        if pq_metadata:
            result["pq_metadata"] = pq_metadata

        return result

    @property
    def row_groups_to_download(self) -> List[int]:
        return self["row_groups_to_download"]

    @property
    def num_row_groups(self) -> List[int]:
        return self["num_row_groups"]

    @property
    def num_rows(self) -> int:
        return self["num_rows"]

    @property
    def in_memory_size_bytes(self) -> float:
        return self["in_memory_size_bytes"]

    @property
    def pq_metadata(self) -> Optional[FileMetaData]:
        return self.get("pq_metadata")

    @pq_metadata.setter
    def pq_metadata(self, metadata: FileMetaData) -> None:
        self["pq_metadata"] = metadata



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from contextlib import AbstractContextManager
from types import TracebackType
import ray
import sys
import threading
import time
from typing import Dict, Any, Optional
from dataclasses import dataclass
from deltacat import logs
import logging
from resource import getrusage, RUSAGE_SELF
import platform
import psutil
import schedule
from deltacat.constants import BYTES_PER_GIBIBYTE


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@dataclass
class ClusterUtilization:
    def __init__(
        self, cluster_resources: Dict[str, Any], available_resources: Dict[str, Any]
    ):
        used_resources = {}

        for key in cluster_resources:
            if (
                isinstance(cluster_resources[key], float)
                or isinstance(cluster_resources[key], int)
            ) and key in available_resources:
                used_resources[key] = cluster_resources[key] - available_resources[key]

        self.total_memory_bytes = cluster_resources.get("memory")
        self.used_memory_bytes = used_resources.get("memory")
        self.total_cpu = cluster_resources.get("CPU")
        self.used_cpu = used_resources.get("CPU")
        self.total_object_store_memory_bytes = cluster_resources.get(
            "object_store_memory"
        )
        self.used_object_store_memory_bytes = used_resources.get("object_store_memory")
        self.used_memory_percent = (
            self.used_memory_bytes / self.total_memory_bytes
        ) * 100
        self.used_object_store_memory_percent = (
            self.used_object_store_memory_bytes / self.total_object_store_memory_bytes
        ) * 100
        self.used_cpu_percent = (self.used_cpu / self.total_cpu) * 100
        self.used_resources = used_resources

    @staticmethod
    def get_current_cluster_utilization() -> ClusterUtilization:
        cluster_resources = ray.cluster_resources()
        available_resources = ray.available_resources()

        return ClusterUtilization(
            cluster_resources=cluster_resources, available_resources=available_resources
        )


class ClusterUtilizationOverTimeRange(AbstractContextManager):
    """
    This class can be used to compute the cluster utilization metrics
    which requires us to compute it over time as they change on-demand.

    For example, in an autoscaling cluster, the vCPUs keep changing and hence
    more important metrics to capture in that scenario is vcpu-seconds.
    """

    def __init__(self) -> None:
        self.total_vcpu_seconds = 0.0
        self.used_vcpu_seconds = 0.0
        self.total_memory_gb_seconds = 0.0
        self.used_memory_gb_seconds = 0.0
        self.max_cpu = 0.0

    def __enter__(self) -> Any:
        schedule.every().second.do(self._update_resources)
        self.stop_run_schedules = self._run_schedule()
        return super().__enter__()

    def __exit__(
        self,
        __exc_type: type[BaseException] | None,
        __exc_value: BaseException | None,
        __traceback: TracebackType | None,
    ) -> bool | None:
        if __exc_value:
            logger.error(
                f"Error ocurred while calculating cluster resources: {__exc_value}"
            )
        self.stop_run_schedules.set()
        return super().__exit__(__exc_type, __exc_value, __traceback)

    # It is not truely parallel(due to GIL Ref: https://wiki.python.org/moin/GlobalInterpreterLock)
    # even if we are using threading library. However, it averages out and gives a very good approximation.
    def _update_resources(self):
        cluster_resources = ray.cluster_resources()
        available_resources = ray.available_resources()
        if "CPU" not in cluster_resources:
            return

        if "CPU" in available_resources:
            self.used_vcpu_seconds = self.used_vcpu_seconds + float(
                str(cluster_resources["CPU"] - available_resources["CPU"])
            )

        self.total_vcpu_seconds = self.total_vcpu_seconds + float(
            str(cluster_resources["CPU"])
        )
        self.max_cpu = max(self.max_cpu, float(str(cluster_resources["CPU"])))

        if "memory" not in cluster_resources:
            return

        if "memory" in available_resources:
            self.used_memory_gb_seconds = (
                self.used_memory_gb_seconds
                + float(
                    str(cluster_resources["memory"] - available_resources["memory"])
                )
                / BYTES_PER_GIBIBYTE
            )

        self.total_memory_gb_seconds = (
            self.total_memory_gb_seconds
            + float(str(cluster_resources["memory"])) / BYTES_PER_GIBIBYTE
        )

    def _run_schedule(self, interval: Optional[float] = 1.0):
        cease_continuous_run = threading.Event()

        class ScheduleThread(threading.Thread):
            @classmethod
            def run(cls):
                while not cease_continuous_run.is_set():
                    schedule.run_pending()
                    time.sleep(float(str(interval)))

        continuous_thread = ScheduleThread()
        continuous_thread.start()
        return cease_continuous_run


def get_current_node_peak_memory_usage_in_bytes():
    """
    Returns the peak memory usage of the node in bytes. This method works across
    Windows, Darwin and Linux platforms.
    """
    current_platform = platform.system()
    if current_platform != "Windows":
        usage = getrusage(RUSAGE_SELF).ru_maxrss
        if current_platform == "Linux":
            usage = usage * 1024
        return usage
    else:
        return psutil.Process().memory_info().peak_wset


def get_size_of_object_in_bytes(obj: object) -> float:
    size = sys.getsizeof(obj)
    if isinstance(obj, dict):
        return (
            size
            + sum(map(get_size_of_object_in_bytes, obj.keys()))
            + sum(map(get_size_of_object_in_bytes, obj.values()))
        )
    if isinstance(obj, (list, tuple, set, frozenset)):
        return size + sum(map(get_size_of_object_in_bytes, obj))
    return size



import time
from collections import Counter
from typing import Any, Callable, Tuple, TypeVar

T = TypeVar("T")


def invoke_with_perf_counter(
    counter: Counter, counter_key: Any, func: Callable, *args, **kwargs
) -> Tuple[Any, float]:

    start = time.perf_counter()
    result = func(*args, **kwargs)
    stop = time.perf_counter()
    latency = stop - start
    counter[counter_key] += latency
    return result, latency


def timed_invocation(func: Callable[[Any], T], *args, **kwargs) -> Tuple[T, float]:

    start = time.perf_counter()
    result = func(*args, **kwargs)
    stop = time.perf_counter()
    return result, stop - start



import ray
import time
import logging
from deltacat import logs
from deltacat.utils.common import ReadKwargsProvider
import functools
import itertools
from deltacat.compute.compactor import (
    RoundCompletionInfo,
)
from deltacat.storage.model.sort_key import SortKey
from deltacat.types.media import ContentType
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.utils.ray_utils.concurrency import (
    invoke_parallel,
    round_robin_options_provider,
)

from deltacat.compute.compactor.model.repartition_result import RepartitionResult
from deltacat.utils.placement import PlacementGroupConfig
from typing import List, Optional, Dict, Any
from deltacat.utils.ray_utils.runtime import live_node_resource_keys
from deltacat.compute.compactor.utils import io
from deltacat.compute.compactor.utils import round_completion_file as rcf
from deltacat.compute.compactor.steps import repartition as repar
from deltacat.compute.compactor.steps.repartition import RepartitionType
from deltacat.storage import (
    Delta,
    DeltaLocator,
    PartitionLocator,
    interface as unimplemented_deltacat_storage,
)
from deltacat.utils.metrics import MetricsConfig
from deltacat.compute.compactor.utils.sort_key import validate_sort_keys

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


# TODO: move this repartition function to a separate module under compute
def repartition(
    source_partition_locator: PartitionLocator,
    destination_partition_locator: PartitionLocator,
    repartition_args: Any,
    repartition_completion_file_s3_url: str,
    last_stream_position_to_compact: int,
    repartition_type: RepartitionType = RepartitionType.RANGE,
    sort_keys: List[SortKey] = None,
    records_per_repartitioned_file: int = 4_000_000,
    min_file_count: int = 1000,
    min_delta_bytes: int = 200 * 2**20,
    repartitioned_file_content_type: ContentType = ContentType.PARQUET,
    enable_profiler: bool = False,
    metrics_config: Optional[MetricsConfig] = None,
    pg_config: Optional[PlacementGroupConfig] = None,
    list_deltas_kwargs: Optional[Dict[str, Any]] = None,
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    s3_client_kwargs: Optional[Dict[str, Any]] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    **kwargs,
) -> Optional[str]:

    node_resource_keys = None
    if pg_config:  # use resource in each placement group
        cluster_resources = pg_config.resource
        cluster_cpus = cluster_resources["CPU"]
    else:  # use all cluster resource
        cluster_resources = ray.cluster_resources()
        logger.info(f"Total cluster resources: {cluster_resources}")
        logger.info(f"Available cluster resources: {ray.available_resources()}")
        cluster_cpus = int(cluster_resources["CPU"])
        logger.info(f"Total cluster CPUs: {cluster_cpus}")
        node_resource_keys = live_node_resource_keys()
        logger.info(
            f"Found {len(node_resource_keys)} live cluster nodes: "
            f"{node_resource_keys}"
        )

    # create a remote options provider to round-robin tasks across all nodes or allocated bundles
    logger.info(f"Setting round robin scheduling with node id:{node_resource_keys}")
    round_robin_opt_provider = functools.partial(
        round_robin_options_provider,
        resource_keys=node_resource_keys,
        pg_config=pg_config.opts if pg_config else None,
    )

    deltas = io._discover_deltas(
        source_partition_locator,
        None,
        deltacat_storage.get_partition(
            source_partition_locator.stream_locator,
            source_partition_locator.partition_values,
        ).stream_position,
        deltacat_storage,
        list_deltas_kwargs,
    )

    uniform_deltas = []
    for delta in deltas:
        uniform_deltas_part = DeltaAnnotated.rebatch(
            [DeltaAnnotated.of(delta)],
            min_delta_bytes=min_delta_bytes,
            min_file_counts=min_file_count,
        )
        uniform_deltas.extend(uniform_deltas_part)

    logger.info(f"Retrieved a total of {len(uniform_deltas)} uniform deltas.")

    max_parallelism = cluster_cpus
    # create a new stream for this round
    compacted_stream_locator = destination_partition_locator.stream_locator
    stream = deltacat_storage.get_stream(
        compacted_stream_locator.namespace,
        compacted_stream_locator.table_name,
        compacted_stream_locator.table_version,
    )
    partition = deltacat_storage.stage_partition(
        stream,
        destination_partition_locator.partition_values,
    )
    new_compacted_partition_locator = partition.locator
    repar_start = time.time()
    repar_tasks_pending = invoke_parallel(
        items=uniform_deltas,
        ray_task=repar.repartition,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        repartition_type=repartition_type,
        repartition_args=repartition_args,
        max_records_per_output_file=records_per_repartitioned_file,
        destination_partition=partition,
        enable_profiler=enable_profiler,
        metrics_config=metrics_config,
        read_kwargs_provider=read_kwargs_provider,
        s3_table_writer_kwargs=s3_table_writer_kwargs,
        repartitioned_file_content_type=repartitioned_file_content_type,
        deltacat_storage=deltacat_storage,
    )
    logger.info(f"Getting {len(repar_tasks_pending)} task results...")
    repar_results: List[RepartitionResult] = ray.get(repar_tasks_pending)
    repar_results: List[Delta] = [rp.range_deltas for rp in repar_results]
    transposed = list(itertools.zip_longest(*repar_results, fillvalue=None))
    ordered_deltas: List[Delta] = [
        i for sublist in transposed for i in sublist if i is not None
    ]
    repar_end = time.time()
    logger.info(f"repartition {repar_end - repar_start} seconds")
    logger.info(f"Got {len(ordered_deltas)} task results.")
    # ordered_deltas are ordered as [cold1, cold2, coldN, hot1, hot2, hotN]
    merged_delta = Delta.merge_deltas(
        ordered_deltas, stream_position=last_stream_position_to_compact
    )
    compacted_delta = deltacat_storage.commit_delta(
        merged_delta, properties=kwargs.get("properties", {})
    )
    deltacat_storage.commit_partition(partition)
    logger.info(f"Committed final delta: {compacted_delta}")
    logger.info(f"Job run completed successfully!")
    new_compacted_delta_locator = DeltaLocator.of(
        new_compacted_partition_locator,
        compacted_delta.stream_position,
    )
    bit_width_of_sort_keys = validate_sort_keys(
        source_partition_locator,
        sort_keys,
        deltacat_storage,
        deltacat_storage_kwargs={},
    )
    repartition_completion_info = RoundCompletionInfo.of(
        last_stream_position_to_compact,
        new_compacted_delta_locator,
        None,
        bit_width_of_sort_keys,
        None,
    )
    if s3_client_kwargs is None:
        s3_client_kwargs = {}

    return rcf.write_round_completion_file(
        None,
        None,
        repartition_completion_info,
        repartition_completion_file_s3_url,
        **s3_client_kwargs,
    )



import pyarrow as pa
from typing import List
from deltacat.storage import PartitionLocator, SortKey

MAX_SORT_KEYS_BIT_WIDTH = 256


def validate_sort_keys(
    source_partition_locator: PartitionLocator,
    sort_keys: List[SortKey],
    deltacat_storage,
    deltacat_storage_kwargs,
    **kwargs,
) -> int:
    """
    Validates the input sort keys to ensure that they are unique, are using
    a valid sort key model, are all fixed-width data types, and that the
    sum of bit widths across sort key data types is less-than-or-equal-to
    256. Returns the sum of bit widths across all sort keys.
    """
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    total_sort_keys_bit_width = 0
    if sort_keys:
        sort_key_names = [key.key_name for key in sort_keys]
        assert len(sort_key_names) == len(
            set(sort_key_names)
        ), f"Sort key names must be unique: {sort_key_names}"
        stream_locator = source_partition_locator.stream_locator
        table_version_schema = deltacat_storage.get_table_version_schema(
            stream_locator.namespace,
            stream_locator.table_name,
            stream_locator.table_version,
            **deltacat_storage_kwargs,
        )
        if isinstance(table_version_schema, pa.Schema):
            for sort_key_name in sort_key_names:
                pa_field: pa.Field = pa.Schema.field(sort_key_name)
                pa_type: pa.DataType = pa_field.type
                try:
                    total_sort_keys_bit_width += pa_type.bit_width
                    if total_sort_keys_bit_width > MAX_SORT_KEYS_BIT_WIDTH:
                        raise ValueError(
                            f"Total length of sort keys "
                            f"({total_sort_keys_bit_width}) is greater "
                            f"than the max supported bit width for all "
                            f"sort keys ({MAX_SORT_KEYS_BIT_WIDTH})"
                        )
                except ValueError as e:
                    raise ValueError(
                        f"Unable to get bit width of sort key: {pa_field}. "
                        f"Please ensure that all sort keys are fixed-size "
                        f"PyArrow data types."
                    ) from e
        else:
            total_sort_keys_bit_width = MAX_SORT_KEYS_BIT_WIDTH
    return total_sort_keys_bit_width



import unittest
from deltacat.types.media import ContentEncoding, ContentType
from deltacat.utils.daft import daft_s3_file_to_table

from deltacat.utils.pyarrow import ReadKwargsProviderPyArrowSchemaOverride
from deltacat.types.partial_download import PartialParquetParameters
import pyarrow as pa

from pyarrow import parquet as pq


class TestDaftParquetReader(unittest.TestCase):
    MVP_PATH = "deltacat/tests/utils/data/mvp.parquet"

    def test_read_from_s3_all_columns(self):
        table = daft_s3_file_to_table(
            self.MVP_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
        )
        self.assertEqual(table.schema.names, ["a", "b"])
        self.assertEqual(table.num_rows, 100)

    def test_read_from_s3_single_column_via_include_columns(self):
        table = daft_s3_file_to_table(
            self.MVP_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            include_columns=["b"],
        )
        self.assertEqual(table.schema.names, ["b"])
        self.assertEqual(table.num_rows, 100)

    def test_read_from_s3_single_column_via_column_names(self):
        table = daft_s3_file_to_table(
            self.MVP_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            column_names=["b"],
        )
        self.assertEqual(table.schema.names, ["b"])
        self.assertEqual(table.num_rows, 100)

    def test_read_from_s3_single_column_with_schema(self):
        schema = pa.schema([("a", pa.int8()), ("b", pa.string())])
        pa_read_func_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(
            schema=schema
        )
        table = daft_s3_file_to_table(
            self.MVP_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            include_columns=["a"],
            pa_read_func_kwargs_provider=pa_read_func_kwargs_provider,
        )
        self.assertEqual(table.schema.names, ["a"])
        self.assertEqual(table.schema.field("a").type, pa.int8())
        self.assertEqual(table.num_rows, 100)

    def test_read_from_s3_single_column_with_schema_reverse_order(self):
        schema = pa.schema([("b", pa.string()), ("a", pa.int8())])
        pa_read_func_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(
            schema=schema
        )
        table = daft_s3_file_to_table(
            self.MVP_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            pa_read_func_kwargs_provider=pa_read_func_kwargs_provider,
        )
        self.assertEqual(table.schema.names, ["b", "a"])
        self.assertEqual(table.schema.field("a").type, pa.int8())
        self.assertEqual(table.num_rows, 100)

    def test_read_from_s3_single_column_with_schema_subset_cols(self):
        schema = pa.schema([("a", pa.int8())])
        pa_read_func_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(
            schema=schema
        )
        table = daft_s3_file_to_table(
            self.MVP_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            pa_read_func_kwargs_provider=pa_read_func_kwargs_provider,
        )
        self.assertEqual(table.schema.names, ["a"])
        self.assertEqual(table.schema.field("a").type, pa.int8())
        self.assertEqual(table.num_rows, 100)

    def test_read_from_s3_single_column_with_schema_extra_cols(self):
        schema = pa.schema([("a", pa.int8()), ("MISSING", pa.string())])
        pa_read_func_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(
            schema=schema
        )
        table = daft_s3_file_to_table(
            self.MVP_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            pa_read_func_kwargs_provider=pa_read_func_kwargs_provider,
        )
        self.assertEqual(
            table.schema.names, ["a", "MISSING"]
        )  # NOTE: "MISSING" is padded as a null array
        self.assertEqual(table.schema.field("a").type, pa.int8())
        self.assertEqual(table.schema.field("MISSING").type, pa.string())
        self.assertEqual(table.num_rows, 100)

    def test_read_from_s3_single_column_with_row_groups(self):

        metadata = pq.read_metadata(self.MVP_PATH)
        ppp = PartialParquetParameters.of(pq_metadata=metadata)
        ppp["row_groups_to_download"] = ppp.row_groups_to_download[1:2]
        table = daft_s3_file_to_table(
            self.MVP_PATH,
            content_encoding=ContentEncoding.IDENTITY.value,
            content_type=ContentType.PARQUET.value,
            column_names=["b"],
            partial_file_download_params=ppp,
        )
        self.assertEqual(table.schema.names, ["b"])
        self.assertEqual(table.num_rows, 10)


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional

import numpy as np
import pyarrow as pa
from fsspec import AbstractFileSystem
from ray.data.datasource import BlockWritePathProvider

from deltacat.types.media import ContentType
from deltacat.utils import pandas as pd_utils
from deltacat.utils import pyarrow as pa_utils
from deltacat.utils.common import ReadKwargsProvider


def slice_ndarray(np_array: np.ndarray, max_len: Optional[int]) -> List[np.ndarray]:
    """
    Iteratively creates max_len slices from the first dimension of an ndarray.
    """
    if max_len is None:
        return [np_array]

    # Slice along the first dimension of the ndarray.
    return [np_array[i : i + max_len] for i in range(0, len(np_array), max_len)]


def s3_file_to_ndarray(
    s3_url: str,
    content_type: str,
    content_encoding: str,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pd_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    **s3_client_kwargs
) -> np.ndarray:
    # TODO: Compare perf to s3 -> pyarrow -> pandas [Series/DataFrame] -> numpy
    dataframe = pd_utils.s3_file_to_dataframe(
        s3_url,
        content_type,
        content_encoding,
        column_names,
        include_columns,
        pd_read_func_kwargs_provider,
        **s3_client_kwargs
    )
    return dataframe.to_numpy()


def ndarray_size(np_array: np.ndarray) -> int:
    return np_array.nbytes


def ndarray_to_file(
    np_array: np.ndarray,
    path: str,
    file_system: AbstractFileSystem,
    block_path_provider: BlockWritePathProvider,
    content_type: str = ContentType.PARQUET.value,
    **kwargs
) -> None:
    """
    Writes the given Numpy ndarray to a file.
    """

    # PyArrow only supports 1D ndarrays, so convert to list of 1D arrays
    np_arrays = [array for array in np_array]
    pa_utils.table_to_file(
        pa.table({"data": np_arrays}),
        path,
        file_system,
        block_path_provider,
        content_type,
        **kwargs
    )



import unittest
from unittest import mock
from deltacat.tests.test_utils.constants import TEST_DELTA
from typing import Any, Dict

DATABASE_FILE_PATH_KEY, DATABASE_FILE_PATH_VALUE = (
    "db_file_path",
    "deltacat/tests/local_deltacat_storage/db_test.sqlite",
)


class TestFitInputDeltas(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.module_patcher = mock.patch.dict("sys.modules", {"ray": mock.MagicMock()})
        cls.module_patcher.start()

        from deltacat.compute.compactor.model.compaction_session_audit_info import (
            CompactionSessionAuditInfo,
        )

        cls.kwargs_for_local_deltacat_storage: Dict[str, Any] = {
            DATABASE_FILE_PATH_KEY: DATABASE_FILE_PATH_VALUE,
        }

        cls.COMPACTION_AUDIT = CompactionSessionAuditInfo("1.0", "2.3", "test")

        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.module_patcher.stop()

    def test_sanity(self):
        from deltacat.compute.compactor.utils import io
        import deltacat.tests.local_deltacat_storage as ds

        (
            delta_list,
            hash_bucket_count,
            high_watermark,
            require_multiple_rounds,
        ) = io.fit_input_deltas(
            [TEST_DELTA],
            {"CPU": 1, "memory": 20000000},
            self.COMPACTION_AUDIT,
            None,
            ds,
            self.kwargs_for_local_deltacat_storage,
        )

        self.assertIsNotNone(hash_bucket_count)
        self.assertTrue(1, len(delta_list))
        self.assertIsNotNone(high_watermark)
        self.assertFalse(require_multiple_rounds)
        self.assertIsNotNone(hash_bucket_count, self.COMPACTION_AUDIT.hash_bucket_count)
        self.assertIsNotNone(self.COMPACTION_AUDIT.input_file_count)
        self.assertIsNotNone(self.COMPACTION_AUDIT.input_size_bytes)
        self.assertIsNotNone(self.COMPACTION_AUDIT.total_cluster_memory_bytes)

    def test_when_hash_bucket_count_overridden(self):
        from deltacat.compute.compactor.utils import io
        import deltacat.tests.local_deltacat_storage as ds

        (
            delta_list,
            hash_bucket_count,
            high_watermark,
            require_multiple_rounds,
        ) = io.fit_input_deltas(
            [TEST_DELTA],
            {"CPU": 1, "memory": 20000000},
            self.COMPACTION_AUDIT,
            20,
            ds,
            self.kwargs_for_local_deltacat_storage,
        )

        self.assertEqual(20, hash_bucket_count)
        self.assertEqual(1, len(delta_list))
        self.assertIsNotNone(high_watermark)
        self.assertFalse(require_multiple_rounds)

    def test_when_not_enough_memory_splits_manifest_entries(self):
        from deltacat.compute.compactor.utils import io
        import deltacat.tests.local_deltacat_storage as ds

        (
            delta_list,
            hash_bucket_count,
            high_watermark,
            require_multiple_rounds,
        ) = io.fit_input_deltas(
            [TEST_DELTA],
            {"CPU": 2, "memory": 10},
            self.COMPACTION_AUDIT,
            20,
            ds,
            self.kwargs_for_local_deltacat_storage,
        )

        self.assertIsNotNone(hash_bucket_count)
        self.assertTrue(2, len(delta_list))
        self.assertIsNotNone(high_watermark)
        self.assertFalse(require_multiple_rounds)

    def test_when_no_input_deltas(self):
        from deltacat.compute.compactor.utils import io
        import deltacat.tests.local_deltacat_storage as ds

        with self.assertRaises(AssertionError):
            io.fit_input_deltas(
                [],
                {"CPU": 100, "memory": 20000.0},
                self.COMPACTION_AUDIT,
                None,
                ds,
                self.kwargs_for_local_deltacat_storage,
            )

    def test_when_cpu_resources_is_not_passed(self):
        from deltacat.compute.compactor.utils import io
        import deltacat.tests.local_deltacat_storage as ds

        with self.assertRaises(KeyError):
            io.fit_input_deltas(
                [],
                {},
                self.COMPACTION_AUDIT,
                None,
                ds,
                self.kwargs_for_local_deltacat_storage,
            )



from typing import List, Any


class IObjectStore:
    """
    An object store interface.
    """

    def setup(self, *args, **kwargs) -> Any:
        ...

        """
        Sets up everything needed to run the object store.
        """

    def put(self, obj: object, *args, **kwargs) -> Any:
        """
        Put a single object into the object store. Depending
        on the implementation, this method can be sync or async.
        """
        return self.put_many([obj])[0]

    def put_many(self, objects: List[object], *args, **kwargs) -> List[Any]:
        ...

        """
        Put many objects into the object store. It would return an ordered list
        of object references corresponding to each object in the input.
        """

    def get(self, ref: Any, *args, **kwargs) -> object:
        """
        Get a single object from an object store.
        """
        return self.get_many([ref])[0]

    def get_many(self, refs: List[Any], *args, **kwargs) -> List[object]:
        ...

        """
        Get a list of objects from the object store. Use this method to
        avoid multiple get calls. Note that depending on implementation it may
        or may not return ordered results.
        """

    def clear(self, *args, **kwargs) -> bool:
        ...

        """
        Clears the object store and all the associated data in it.
        """

    def close(self, *args, **kwargs) -> None:
        ...
        """
        Closes all the active connections to object store without clearing
        the data in the object store.
        """



import json
import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional

import pyarrow
import ray

from deltacat import LocalTable, TableType, logs
from deltacat.aws import s3u as s3_utils
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.compute.metastats.model.partition_stats_dict import PartitionStats
from deltacat.compute.stats.models.delta_column_stats import DeltaColumnStats
from deltacat.compute.stats.models.delta_stats import DeltaStats, DeltaStatsCacheMiss
from deltacat.compute.stats.models.delta_stats_cache_result import DeltaStatsCacheResult
from deltacat.compute.stats.models.stats_result import StatsResult
from deltacat.storage import Delta
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.utils.common import sha1_hexdigest

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def cache_inflation_rate_data_for_delta_stats_ready(
    delta_stats_processed_list, inflation_rate_stats_s3_url, deltacat_storage
):
    meta_stats_processed_list: Dict[int, int] = {}

    for key, value in delta_stats_processed_list.items():
        delta_locator = value.column_stats[0].manifest_stats.delta_locator
        delta_meta_count = 0
        manifest = deltacat_storage.get_delta_manifest(delta_locator)
        delta = Delta.of(delta_locator, None, None, None, manifest)
        for entry in delta.manifest.entries:
            delta_meta_count += entry.meta.content_length
        meta_stats_processed_list[delta.stream_position] = delta_meta_count

    cache_inflation_rate_res = dict()

    for key, value in delta_stats_processed_list.items():
        delta_stats_pyarrow_bytes_sum = 0
        delta_stats_row_count = 0
        for column_stats in (
            delta_stats_processed_list[key].column_stats[0].manifest_stats.stats
        ):
            delta_stats_row_count += column_stats.get("rowCount")
        for stats in delta_stats_processed_list[key].get("column_stats"):

            delta_stats_pyarrow_bytes_sum += stats.get("stats").get("pyarrowTableBytes")
        cache_inflation_rate_res[key] = [
            meta_stats_processed_list[key],
            delta_stats_row_count,
            delta_stats_pyarrow_bytes_sum,
        ]

    if inflation_rate_stats_s3_url:
        logger.warning(
            f"reading previous inflation rate stats from: {inflation_rate_stats_s3_url}"
        )

        result = s3_utils.download(inflation_rate_stats_s3_url, fail_if_not_found=False)

        prev_inflation_rate_stats = dict()
        if result:
            json_str = result["Body"].read().decode("utf-8")
            prev_inflation_rate_stats_read = json.loads(json_str)
            prev_inflation_rate_stats = (
                prev_inflation_rate_stats_read
                if prev_inflation_rate_stats_read
                else dict()
            )
            logger.debug(
                f"read stats completion info: {prev_inflation_rate_stats_read}"
            )
        logger.debug(
            f"writing inflation rate info to S3: {inflation_rate_stats_s3_url}"
        )
        prev_inflation_rate_stats.update(cache_inflation_rate_res)
        logger.debug(
            f"writing current inflation rate info to S3: {prev_inflation_rate_stats}"
        )
        s3_utils.upload(
            inflation_rate_stats_s3_url, json.dumps(prev_inflation_rate_stats)
        )
    else:
        logger.warning(
            f"No valid s3 url received to cache inflation rate stats, got {inflation_rate_stats_s3_url}"
        )


def read_cached_partition_stats(
    partition_canonical_string: str, stat_results_s3_bucket: str
):
    partition_stats_url = get_partition_stats_s3_url(
        partition_canonical_string, stat_results_s3_bucket
    )
    logger.info(f"reading partition stats completion file from: {partition_stats_url}")

    result = s3_utils.download(partition_stats_url, fail_if_not_found=False)
    delta_stats_cache_res_map: Dict[int, List[DeltaStatsCacheResult]] = {}
    if result:
        json_str = result["Body"].read().decode("utf-8")
        partition_stats_str = json.loads(json_str)
        delta_stats_cache_res_map = get_delta_stats_from_partition_stats(
            partition_stats_str
        )

    return delta_stats_cache_res_map


def get_partition_stats_s3_url(
    partition_canonical_string: str, stat_results_s3_bucket: str
):
    stats_partition_canonical_string = f"{partition_canonical_string}"
    stats_partition_hexdigest = sha1_hexdigest(
        stats_partition_canonical_string.encode("utf-8")
    )
    base_path = s3_utils.parse_s3_url(stat_results_s3_bucket).url

    return f"{base_path}/{stats_partition_hexdigest}.json"


def get_delta_stats_from_partition_stats(partition_stats_str: str):

    partition_stats = PartitionStats.build_from_dict(partition_stats_str)

    found_columns_stats_map: Dict[int, List[DeltaStatsCacheResult]] = {}
    for stream_position, delta_stats in partition_stats.delta_stats.items():
        found_columns_stats: List[DeltaColumnStats] = []
        missed_columns: List[str] = []
        for cs in delta_stats.column_stats:
            if cs.manifest_stats:
                found_columns_stats.append(cs)
            else:
                missed_columns.append(cs.column)

        delta_locator = delta_stats.column_stats[0].manifest_stats.delta_locator
        found_stats: Optional[DeltaStats] = (
            DeltaStats.of(found_columns_stats) if found_columns_stats else None
        )
        missed_stats: Optional[DeltaStatsCacheMiss] = (
            DeltaStatsCacheMiss(missed_columns, delta_locator)
            if missed_columns
            else None
        )
        delta_stats_cache_res = DeltaStatsCacheResult.of(found_stats, missed_stats)
        found_columns_stats_map[int(stream_position)] = delta_stats_cache_res
    return found_columns_stats_map


def cache_partition_stats_to_s3(
    stat_results_s3_bucket, delta_stream_range_stats, partition_canonical_string
):
    partition_stats = PartitionStats.of(
        delta_stream_range_stats, partition_canonical_string
    )
    logger.info(f"writing partition stats completion for {partition_canonical_string}")
    partition_stats_completion_file_s3_url = get_partition_stats_s3_url(
        partition_canonical_string, stat_results_s3_bucket
    )
    s3_utils.upload(
        partition_stats_completion_file_s3_url, str(json.dumps(partition_stats))
    )
    logger.debug(
        f"stats completion file written to: {partition_stats_completion_file_s3_url}"
    )


@ray.remote
def collect_stats_by_columns(
    delta_annotated: DeltaAnnotated,
    columns_to_compute: Optional[List[str]] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Materializes one manifest entry at a time to save memory usage and calculate stats from each of its columns.

    Args:
        delta: A delta object to calculate stats for
        columns_to_compute: Columns to calculate stats for. If not provided, all columns are considered.
        deltacat_storage: Client implementation of the DeltaCAT storage interface

    Returns:
        A delta wide stats container
    """
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    total_tables_size = 0

    # Mapping of column_name -> [stats_file_idx_1, stats_file_idx_2, ... stats_file_idx_n]
    column_stats_map = defaultdict(
        lambda: [[None, None]] * len(delta_annotated["manifest"].get("entries"))
    )
    src_da_entries = delta_annotated["manifest"].get("entries")
    manifest_annotations = delta_annotated["annotations"]
    for file_idx, manifest in enumerate(src_da_entries):
        entry_pyarrow_table: LocalTable = (
            deltacat_storage.download_delta_manifest_entry(
                delta_annotated,
                file_idx,
                TableType.PYARROW,
                columns_to_compute,
                equivalent_table_types="uncompacted",
                **deltacat_storage_kwargs,
            )
        )
        assert isinstance(entry_pyarrow_table, pyarrow.Table), (
            f"Stats collection is only supported for PyArrow tables, but received a table of "
            f"type '{type(entry_pyarrow_table)}' for manifest entry {file_idx} of delta: {delta_annotated.locator}."
        )
        total_tables_size += entry_pyarrow_table.nbytes
        if not columns_to_compute:
            columns_to_compute = entry_pyarrow_table.column_names

        for column_idx, pyarrow_column in enumerate(entry_pyarrow_table.columns):
            column_name = columns_to_compute[column_idx]
            origin_delta_stream_position = manifest_annotations[file_idx][-1]
            column_stats_map[column_name][file_idx] = [
                StatsResult.of(len(pyarrow_column), pyarrow_column.nbytes),
                origin_delta_stream_position,
            ]

    return column_stats_map



# Allow self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import ray

from deltacat.catalog import interface as catalog_interface

all_catalogs: Optional[Catalogs] = None


class Catalog:
    def __init__(self, impl=catalog_interface, *args, **kwargs):
        self._impl = impl
        self._impl.initialize(*args, **kwargs)

    @property
    def impl(self):
        return self._impl


@ray.remote
class Catalogs:
    def __init__(
        self,
        catalogs: Dict[str, Catalog],
        default_catalog_name: str = None,
        *args,
        **kwargs,
    ):
        if default_catalog_name and default_catalog_name not in catalogs:
            raise ValueError(
                f"Catalog {default_catalog_name} not found "
                f"in catalogs to register: {catalogs}"
            )
        if not catalogs:
            raise ValueError(
                f"No catalogs given to register. "
                f"Please specify one or more catalogs."
            )
        self.catalogs: Dict[str, Catalog] = catalogs
        if default_catalog_name:
            self.default_catalog = self.catalogs[default_catalog_name]
        elif len(catalogs) == 1:
            self.default_catalog = list(self.catalogs.values())[0]
        else:
            self.default_catalog = None

    def all(self) -> Dict[str, Catalog]:
        return self.catalogs

    def names(self) -> List[str]:
        return list(self.catalogs.keys())

    def put(self, name: str, catalog: Catalog) -> None:
        self.catalogs[name] = catalog

    def get(self, name) -> Catalog:
        return self.catalogs.get(name)

    def default(self) -> Optional[Catalog]:
        return self.default_catalog


def init(
    catalogs: Dict[str, Catalog],
    default_catalog_name: str = None,
    ray_init_args: Dict[str, Any] = None,
    *args,
    **kwargs,
) -> None:

    if not ray.is_initialized():
        if ray_init_args:
            ray.init(**ray_init_args)
        else:
            ray.init(address="auto")

    global all_catalogs
    all_catalogs = Catalogs.remote(
        catalogs=catalogs, default_catalog_name=default_catalog_name
    )



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Callable, Generic, List, Optional, TypeVar

import ray

T = TypeVar("T")


class ListResult(dict, Generic[T]):
    @staticmethod
    def of(
        items: Optional[List[T]],
        pagination_key: Optional[str],
        next_page_provider: Optional[Callable[..., ListResult[T]]],
    ) -> ListResult:
        list_result = ListResult()
        list_result["items"] = items
        list_result["paginationKey"] = pagination_key
        list_result["nextPageProvider"] = next_page_provider
        return list_result

    def read_page(self) -> Optional[List[T]]:
        return self.get("items")

    @property
    def pagination_key(self) -> Optional[str]:
        return self.get("paginationKey")

    @property
    def next_page_provider(self) -> Optional[Callable[..., ListResult[T]]]:
        return self.get("nextPageProvider")

    def next_page(self) -> Optional[ListResult[T]]:
        pagination_key = self.pagination_key
        if pagination_key:
            next_page_provider = self.next_page_provider
            if next_page_provider is None:
                raise ValueError(
                    f"Pagination key ('{pagination_key}') "
                    f"specified without a next page provider!"
                )
            next_list_result = next_page_provider(pagination_key)
            if next_list_result.next_page_provider is None:
                next_list_result["nextPageProvider"] = next_page_provider
            return next_list_result
        return None

    def all_items(self) -> List[T]:
        """
        Eagerly loads and returns a flattened list of all pages of items from
        this list result.

        Returns:
            items (List[Any]): Flattened list of items from all list result
                pages.
        """
        items = []
        list_result = self
        while list_result:
            items.extend(list_result.read_page())
            list_result = list_result.next_page()
        return items

    @staticmethod
    @ray.remote
    def all_items_ray(list_result: ListResult) -> List[T]:
        """
        Eagerly loads and returns a flattened list of all pages of items from
        this list result.

        Returns:
            items (List[Any]): Flattened list of items from all list result
                pages.
        """
        return list_result.all_items()



from typing import Dict, List, Optional, Set, Tuple

import ray
from ray.types import ObjectRef

from deltacat.compute.stats.models.delta_stats import DeltaStats
from deltacat.compute.stats.models.delta_stats_cache_result import DeltaStatsCacheResult
from deltacat.compute.stats.models.stats_result import StatsResult
from deltacat.compute.stats.types import StatsType
from deltacat.compute.stats.utils.intervals import DeltaRange, merge_intervals
from deltacat.compute.stats.utils.io import (
    cache_delta_column_stats,
    get_delta_stats,
    get_deltas_from_range,
    read_cached_delta_stats,
)
from deltacat.storage import Delta, DeltaLocator, PartitionLocator
from deltacat.storage import interface as unimplemented_deltacat_storage

# TODO (ricmiyam): Decouple DeltaCAT from S3-based paths
# TODO (ricmiyam): Determine cache eviction policy


def collect(
    source_partition_locator: PartitionLocator,
    delta_stream_position_range_set: Optional[Set[DeltaRange]] = None,
    columns: Optional[List[str]] = None,
    stat_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> Dict[int, DeltaStats]:
    """Collects statistics on deltas, given a set of delta stream position ranges.

    Example:
        >>> collect(locator, set((1, 5), (4, 8), (13, 16)))
        {
            1: DeltaStats(),  # DeltaStats for stream positions 1 - 8
            13: DeltaStats()  # DeltaStats for stream positions 13 - 16
        }

    Args:
        source_partition_locator: Reference to the partition locator tied to the given delta stream positions
        delta_stream_position_range_set: A set of intervals with an int type representing finite,
            closed bounded values, and a None type representing unbounded infinity.
        columns: Columns can be optionally included to collect stats on specific columns.
            By default, all columns will be calculated.
        stat_results_s3_bucket: Used as a cache file storage for computed delta stats
        deltacat_storage: Client implementation of the DeltaCAT storage interface

    Returns:
        A mapping of stream positions to their corresponding delta stats.
    """
    if delta_stream_position_range_set is None:
        delta_stream_position_range_set = {(None, None)}
    delta_stream_range_stats: Dict[int, DeltaStats] = {}
    delta_range_lookup_pending: List[ObjectRef[List[Delta]]] = []

    if not columns:
        columns = deltacat_storage.get_table_version_column_names(
            source_partition_locator.namespace,
            source_partition_locator.table_name,
            source_partition_locator.table_version,
        )
    for range_pair in merge_intervals(delta_stream_position_range_set):
        begin, end = range_pair
        promise: ObjectRef[List[Delta]] = get_deltas_from_range.remote(
            source_partition_locator, begin, end, deltacat_storage
        )
        delta_range_lookup_pending.append(promise)

    delta_list_by_ranges: List[List[Delta]] = ray.get(delta_range_lookup_pending)
    deltas = [delta for delta_list in delta_list_by_ranges for delta in delta_list]

    delta_stats_processed_list: List[DeltaStats] = _collect_stats_from_deltas(
        deltas, columns, stat_results_s3_bucket, deltacat_storage
    )

    for delta_column_stats in delta_stats_processed_list:
        assert (
            len(delta_column_stats.column_stats) > 0
        ), f"Expected columns of `{delta_column_stats}` to be non-empty"
        stream_position = delta_column_stats.column_stats[
            0
        ].manifest_stats.delta_locator.stream_position
        delta_stream_range_stats[stream_position] = delta_column_stats

    return delta_stream_range_stats


def collect_from_deltas(
    deltas: List[Delta],
    stat_types: Set[StatsType],
    columns: Optional[List[str]] = None,
    stat_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> StatsResult:
    """
    Variant of the `collect` function that takes a list of deltas and computes
    the aggregate of all the delta stats.
    """
    if columns is None and deltas:
        delta_locator: DeltaLocator = deltas[0].locator
        columns = deltacat_storage.get_table_version_column_names(
            delta_locator.namespace,
            delta_locator.table_name,
            delta_locator.table_version,
        )

    delta_stats_processed_list: List[DeltaStats] = _collect_stats_from_deltas(
        deltas, columns, stat_results_s3_bucket, deltacat_storage
    )

    return StatsResult.merge(
        [delta_ds.stats for delta_ds in delta_stats_processed_list], stat_types
    )


def _collect_stats_from_deltas(
    deltas: List[Delta],
    columns: Optional[List[str]] = None,
    stat_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> List[DeltaStats]:
    delta_cache_lookup_pending: List[ObjectRef[DeltaStatsCacheResult]] = []
    delta_stats_compute_pending: List[ObjectRef[DeltaStats]] = []

    for delta in deltas:
        if stat_results_s3_bucket:
            promise: ObjectRef[DeltaStatsCacheResult] = read_cached_delta_stats.remote(
                delta, columns, stat_results_s3_bucket
            )
            delta_cache_lookup_pending.append(promise)
            continue

        delta_stats_compute_pending.append(
            get_delta_stats.remote(delta.locator, columns, deltacat_storage)
        )

    return _process_stats(
        delta_cache_lookup_pending,
        delta_stats_compute_pending,
        stat_results_s3_bucket,
        deltacat_storage,
    )


def _process_stats(
    delta_cache_lookup_pending: List[ObjectRef[DeltaStatsCacheResult]],
    delta_stats_compute_pending: List[ObjectRef[DeltaStats]],
    stat_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> List[DeltaStats]:
    if stat_results_s3_bucket:
        delta_stats_processed_list: List[DeltaStats] = _resolve_pending_stats_and_cache(
            delta_cache_lookup_pending, stat_results_s3_bucket, deltacat_storage
        )
    else:
        delta_stats_processed_list: List[DeltaStats] = _resolve_pending_stats(
            delta_stats_compute_pending
        )

    return delta_stats_processed_list


def _resolve_pending_stats_and_cache(
    delta_cache_lookup_pending: List[ObjectRef[DeltaStatsCacheResult]],
    stat_results_s3_bucket: str,
    deltacat_storage,
) -> List[DeltaStats]:
    delta_stats_cached_list, delta_stats_pending_list = _get_cached_and_pending_stats(
        delta_cache_lookup_pending, deltacat_storage
    )
    delta_stats_resolved_list: List[DeltaStats] = _resolve_pending_stats(
        delta_stats_pending_list
    )

    # Cache the stats into the file store
    delta_stats_to_cache: List[ObjectRef] = [
        cache_delta_column_stats.remote(stat_results_s3_bucket, dcs)
        for dataset_stats in delta_stats_resolved_list
        for dcs in dataset_stats.column_stats
    ]
    ray.get(delta_stats_to_cache)

    return [*delta_stats_cached_list, *delta_stats_resolved_list]


def _get_cached_and_pending_stats(
    discover_deltas_pending: List[ObjectRef[DeltaStatsCacheResult]],
    deltacat_storage=unimplemented_deltacat_storage,
) -> Tuple[List[DeltaStats], List[ObjectRef[DeltaStats]]]:
    """
    Returns a tuple of a list of delta stats fetched from the cache, and a list of Ray tasks which will
    calculate the stats for deltas on cache miss.
    """
    delta_stats_processed: List[DeltaStats] = []
    delta_stats_pending: List[ObjectRef[DeltaStats]] = []
    while discover_deltas_pending:
        ready, discover_deltas_pending = ray.wait(discover_deltas_pending)

        cached_results: List[DeltaStatsCacheResult] = ray.get(ready)
        for cached_result in cached_results:
            if cached_result.hits:
                delta_stats_processed.append(cached_result.hits)

            if cached_result.misses:
                missed_column_names: List[str] = cached_result.misses.column_names
                delta_locator: DeltaLocator = cached_result.misses.delta_locator
                delta_stats_pending.append(
                    get_delta_stats.remote(
                        delta_locator, missed_column_names, deltacat_storage
                    )
                )

    return delta_stats_processed, delta_stats_pending


def _resolve_pending_stats(
    delta_stats_pending_list: List[ObjectRef[DeltaStats]],
) -> List[DeltaStats]:
    delta_stats_processed_list: List[DeltaStats] = []
    while delta_stats_pending_list:
        ready, delta_stats_pending_list = ray.wait(delta_stats_pending_list)
        processed_stats_batch: List[DeltaStats] = ray.get(ready)
        delta_stats_processed_list.extend(processed_stats_batch)

    return delta_stats_processed_list



import logging
from ray import cloudpickle
import time
from deltacat.io.object_store import IObjectStore
from typing import Any, List
from deltacat import logs
import uuid
import socket
import redis
from collections import defaultdict


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class RedisObjectStore(IObjectStore):
    """
    An implementation of object store that uses Redis in memory DB.
    """

    def __init__(self) -> None:
        self.client_cache = {}
        self.current_ip = None
        self.SEPARATOR = "_"
        super().__init__()

    def put(self, obj: object, *args, **kwargs) -> Any:
        serialized = cloudpickle.dumps(obj)
        uid = uuid.uuid4()
        current_ip = self._get_current_ip()
        ref = self._create_ref(uid, current_ip)

        client = self._get_client_by_ip(current_ip)
        if client.set(uid.__str__(), serialized):
            return ref
        else:
            raise RuntimeError(f"Unable to write {ref} to cache")

    def put_many(self, objects: List[object], *args, **kwargs) -> List[Any]:
        input = {}
        result = []
        current_ip = self._get_current_ip()
        for obj in objects:
            serialized = cloudpickle.dumps(obj)
            uid = uuid.uuid4()
            ref = self._create_ref(uid, current_ip)
            input[uid.__str__()] = serialized
            result.append(ref)

        client = self._get_client_by_ip(current_ip)

        if client.mset(input):
            return result
        else:
            raise RuntimeError("Unable to update cache")

    def get_many(self, refs: List[Any], *args, **kwargs) -> List[object]:
        result = []
        uid_per_ip = defaultdict(lambda: [])

        start = time.monotonic()
        for ref in refs:
            uid, ip = ref.split(self.SEPARATOR)
            uid_per_ip[ip].append(uid)

        for (ip, uids) in uid_per_ip.items():
            client = self._get_client_by_ip(ip)
            cache_result = client.mget(uids)
            assert len(cache_result) == len(
                uids
            ), "Not all values were returned from cache"

            total_bytes = 0

            deserialize_start = time.monotonic()
            for serialized in cache_result:
                deserialized = cloudpickle.loads(serialized)
                total_bytes += len(serialized)
                result.append(deserialized)

            deserialize_end = time.monotonic()
            logger.debug(
                f"The time taken to deserialize {total_bytes} bytes is: {deserialize_end - deserialize_start}",
            )

        end = time.monotonic()

        logger.info(f"The total time taken to read all objects is: {end - start}")

        return result

    def get(self, ref: Any, *args, **kwargs) -> object:
        uid, ip = ref.split(self.SEPARATOR)
        client = self._get_client_by_ip(ip)
        serialized = client.get(uid)
        return cloudpickle.loads(serialized)

    def _get_client_by_ip(self, ip_address: str):
        if ip_address in self.client_cache:
            return self.client_cache[ip_address]

        base_client = redis.Redis(ip_address, 7777)

        self.client_cache[ip_address] = base_client
        return base_client

    def _get_current_ip(self):
        if self.current_ip is None:
            self.current_ip = socket.gethostbyname(socket.gethostname())

        return self.current_ip

    def _create_ref(self, uid, ip):
        return f"{uid}{self.SEPARATOR}{ip}"



from enum import Enum
from typing import Dict, Set


class ContentType(str, Enum):
    # See also:
    # https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.17
    # https://www.iana.org/assignments/media-types/media-types.xhtml

    # IANA registered types
    CSV = "text/csv"
    JSON = "application/json"

    # unregistered types
    TSV = "text/tsv"
    PSV = "text/psv"
    PARQUET = "application/parquet"
    ORC = "application/orc"
    FEATHER = "application/feather"
    UNESCAPED_TSV = "application/x-amzn-unescaped-tsv"
    ION = "application/x-amzn-ion"


class ContentEncoding(str, Enum):
    # See also:
    # https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.11
    # http://www.iana.org/assignments/http-parameters/http-parameters.xhtml#content-coding

    # IANA registered encodings
    GZIP = "gzip"
    DEFLATE = "deflate"
    IDENTITY = "identity"
    ZSTD = "zstd"

    # unregistered encodings
    BZIP2 = "bzip2"
    SNAPPY = "snappy"


class TableType(str, Enum):
    PYARROW = "pyarrow"
    PANDAS = "pandas"
    NUMPY = "numpy"
    PYARROW_PARQUET = "pyarrow_parquet"


class SchemaType(str, Enum):
    ARROW = "arrow"


class StorageType(str, Enum):
    LOCAL = "local"
    DISTRIBUTED = "distributed"


DELIMITED_TEXT_CONTENT_TYPES: Set[str] = {
    ContentType.UNESCAPED_TSV.value,
    ContentType.TSV.value,
    ContentType.CSV.value,
    ContentType.PSV.value,
}

TABULAR_CONTENT_TYPES: Set[str] = {
    ContentType.UNESCAPED_TSV.value,
    ContentType.TSV.value,
    ContentType.CSV.value,
    ContentType.PSV.value,
    ContentType.PARQUET.value,
    ContentType.ORC.value,
    ContentType.FEATHER.value,
}

EXPLICIT_COMPRESSION_CONTENT_TYPES: Set[str] = {
    ContentType.UNESCAPED_TSV.value,
    ContentType.TSV.value,
    ContentType.CSV.value,
    ContentType.PSV.value,
    ContentType.JSON.value,
}

CONTENT_TYPE_TO_USER_KWARGS_KEY: Dict[str, str] = {
    ContentType.UNESCAPED_TSV.value: "unescaped_tsv",
    ContentType.TSV.value: "csv",
    ContentType.CSV.value: "csv",
    ContentType.PSV.value: "csv",
    ContentType.PARQUET.value: "parquet",
    ContentType.FEATHER.value: "feather",
    ContentType.ORC.value: "orc",
    ContentType.JSON.value: "json",
}



from __future__ import annotations
import importlib
import copy
import json
from typing import Any, Dict, List, Optional
from deltacat.io.object_store import IObjectStore
from deltacat.utils.common import ReadKwargsProvider
from deltacat.types.media import ContentType
from deltacat.utils.placement import PlacementGroupConfig
from deltacat.io.ray_plasma_object_store import RayPlasmaObjectStore
from deltacat.storage import (
    interface as unimplemented_deltacat_storage,
    PartitionLocator,
    SortKey,
)
from deltacat.compute.compactor_v2.constants import (
    MAX_RECORDS_PER_COMPACTED_FILE,
    MIN_DELTA_BYTES_IN_BATCH,
    MIN_FILES_IN_BATCH,
    AVERAGE_RECORD_SIZE_BYTES,
    TASK_MAX_PARALLELISM,
    DROP_DUPLICATES,
)
from deltacat.constants import PYARROW_INFLATION_MULTIPLIER
from deltacat.compute.compactor.utils.sort_key import validate_sort_keys
from deltacat.utils.metrics import MetricsConfig


class CompactPartitionParams(dict):
    """
    This class represents the parameters passed to compact_partition (deltacat/compute/compactor/compaction_session.py)
    """

    @staticmethod
    def of(params: Optional[Dict]) -> CompactPartitionParams:
        params = {} if params is None else params
        assert (
            params.get("destination_partition_locator") is not None
        ), "destination_partition_locator is a required arg"
        assert (
            params.get("last_stream_position_to_compact") is not None
        ), "last_stream_position_to_compact is a required arg"
        assert (
            params.get("source_partition_locator") is not None
        ), "source_partition_locator is a required arg"
        assert (
            params.get("compaction_artifact_s3_bucket") is not None
        ), "compaction_artifact_s3_bucket is a required arg"

        result = CompactPartitionParams(params)

        # TODO: move defaults to single file
        result.records_per_compacted_file = params.get(
            "records_per_compacted_file", MAX_RECORDS_PER_COMPACTED_FILE
        )
        result.compacted_file_content_type = params.get(
            "compacted_file_content_type", ContentType.PARQUET
        )
        result.object_store = params.get("object_store", RayPlasmaObjectStore())
        result.enable_profiler = params.get("enable_profiler", False)
        result.deltacat_storage = params.get(
            "deltacat_storage", unimplemented_deltacat_storage
        )
        result.s3_client_kwargs = params.get("s3_client_kwargs", {})
        result.deltacat_storage_kwargs = params.get("deltacat_storage_kwargs", {})
        result.list_deltas_kwargs = params.get("list_deltas_kwargs", {})
        result.s3_table_writer_kwargs = params.get("s3_table_writer_kwargs", {})
        result.bit_width_of_sort_keys = validate_sort_keys(
            result.source_partition_locator,
            result.sort_keys,
            result.deltacat_storage,
            result.deltacat_storage_kwargs,
        )
        result.task_max_parallelism = params.get(
            "task_max_parallelism", TASK_MAX_PARALLELISM
        )
        result.min_files_in_batch = params.get("min_files_in_batch", MIN_FILES_IN_BATCH)
        result.min_delta_bytes_in_batch = params.get(
            "min_delta_bytes_in_batch", MIN_DELTA_BYTES_IN_BATCH
        )
        result.previous_inflation = params.get(
            "previous_inflation", PYARROW_INFLATION_MULTIPLIER
        )
        result.average_record_size_bytes = params.get(
            "average_record_size_bytes", AVERAGE_RECORD_SIZE_BYTES
        )
        result.hash_group_count = params.get(
            "hash_group_count", result.hash_bucket_count
        )
        result.drop_duplicates = params.get("drop_duplicates", DROP_DUPLICATES)
        result.ray_custom_resources = params.get("ray_custom_resources")

        result.metrics_config = params.get("metrics_config")

        if not importlib.util.find_spec("memray"):
            result.enable_profiler = False

        if result.primary_keys:
            result.primary_keys = sorted(result.primary_keys)

        # assertions
        assert (
            result.source_partition_locator.partition_values
            == result.destination_partition_locator.partition_values
        ), "Source and destination partitions values must be equal"

        assert (
            result.records_per_compacted_file and result.records_per_compacted_file >= 1
        ), "Max records per output file must be a positive value"

        return result

    @property
    def destination_partition_locator(self) -> PartitionLocator:
        val = self["destination_partition_locator"]
        if not isinstance(val, PartitionLocator):
            val = PartitionLocator(val)

        return val

    @destination_partition_locator.setter
    def destination_partition_locator(self, locator: PartitionLocator) -> None:
        self["destination_partition_locator"] = locator

    @property
    def last_stream_position_to_compact(self) -> int:
        return self["last_stream_position_to_compact"]

    @last_stream_position_to_compact.setter
    def last_stream_position_to_compact(self, stream_position: int) -> None:
        self["last_stream_position_to_compact"] = stream_position

    @property
    def source_partition_locator(self) -> PartitionLocator:
        val = self["source_partition_locator"]
        if not isinstance(val, PartitionLocator):
            val = PartitionLocator(val)
        return val

    @source_partition_locator.setter
    def source_partition_locator(self, locator: PartitionLocator) -> None:
        self["source_partition_locator"] = locator

    @property
    def compaction_artifact_s3_bucket(self) -> str:
        return self["compaction_artifact_s3_bucket"]

    @compaction_artifact_s3_bucket.setter
    def compaction_artifact_s3_bucket(self, s3_bucket: str) -> None:
        self["compaction_artifact_s3_bucket"] = s3_bucket

    @property
    def deltacat_storage(self) -> unimplemented_deltacat_storage:
        return self["deltacat_storage"]

    @deltacat_storage.setter
    def deltacat_storage(self, storage: unimplemented_deltacat_storage) -> None:
        self["deltacat_storage"] = storage

    @property
    def object_store(self) -> IObjectStore:
        return self["object_store"]

    @object_store.setter
    def object_store(self, obj_store: IObjectStore) -> None:
        self["object_store"] = obj_store

    @property
    def compacted_file_content_type(self) -> ContentType:
        return self["compacted_file_content_type"]

    @compacted_file_content_type.setter
    def compacted_file_content_type(self, content_type: ContentType) -> None:
        self["compacted_file_content_type"] = content_type

    @property
    def task_max_parallelism(self) -> int:
        return self["task_max_parallelism"]

    @task_max_parallelism.setter
    def task_max_parallelism(self, max_parallelism: int) -> None:
        self["task_max_parallelism"] = max_parallelism

    @property
    def average_record_size_bytes(self) -> float:
        return self["average_record_size_bytes"]

    @average_record_size_bytes.setter
    def average_record_size_bytes(self, average_record_size_bytes: float) -> None:
        self["average_record_size_bytes"] = average_record_size_bytes

    @property
    def min_files_in_batch(self) -> float:
        return self["min_files_in_batch"]

    @min_files_in_batch.setter
    def min_files_in_batch(self, min_files_in_batch: float) -> None:
        self["min_files_in_batch"] = min_files_in_batch

    @property
    def min_delta_bytes_in_batch(self) -> float:
        return self["min_delta_bytes_in_batch"]

    @min_delta_bytes_in_batch.setter
    def min_delta_bytes_in_batch(self, min_delta_bytes_in_batch: float) -> None:
        self["min_delta_bytes_in_batch"] = min_delta_bytes_in_batch

    @property
    def previous_inflation(self) -> float:
        return self["previous_inflation"]

    @previous_inflation.setter
    def previous_inflation(self, previous_inflation: float) -> None:
        self["previous_inflation"] = previous_inflation

    @property
    def enable_profiler(self) -> bool:
        return self["enable_profiler"]

    @enable_profiler.setter
    def enable_profiler(self, value: bool) -> None:
        self["enable_profiler"] = value

    @property
    def list_deltas_kwargs(self) -> dict:
        return self["list_deltas_kwargs"]

    @list_deltas_kwargs.setter
    def list_deltas_kwargs(self, kwargs: dict) -> None:
        self["list_deltas_kwargs"] = kwargs

    @property
    def s3_table_writer_kwargs(self) -> dict:
        return self["s3_table_writer_kwargs"]

    @s3_table_writer_kwargs.setter
    def s3_table_writer_kwargs(self, kwargs: dict) -> None:
        self["s3_table_writer_kwargs"] = kwargs

    @property
    def deltacat_storage_kwargs(self) -> dict:
        return self["deltacat_storage_kwargs"]

    @deltacat_storage_kwargs.setter
    def deltacat_storage_kwargs(self, kwargs: dict) -> None:
        self["deltacat_storage_kwargs"] = kwargs

    @property
    def s3_client_kwargs(self) -> dict:
        return self["s3_client_kwargs"]

    @s3_client_kwargs.setter
    def s3_client_kwargs(self, kwargs: dict) -> None:
        self["s3_client_kwargs"] = kwargs

    @property
    def records_per_compacted_file(self) -> int:
        return self["records_per_compacted_file"]

    @records_per_compacted_file.setter
    def records_per_compacted_file(self, count: int) -> None:
        self["records_per_compacted_file"] = count

    @property
    def drop_duplicates(self) -> bool:
        return self["drop_duplicates"]

    @drop_duplicates.setter
    def drop_duplicates(self, value: bool):
        self["drop_duplicates"] = value

    @property
    def bit_width_of_sort_keys(self) -> int:
        return self["bit_width_of_sort_keys"]

    @bit_width_of_sort_keys.setter
    def bit_width_of_sort_keys(self, width: int) -> None:
        self["bit_width_of_sort_keys"] = width

    @property
    def hash_bucket_count(self) -> Optional[int]:
        return self.get("hash_bucket_count")

    @hash_bucket_count.setter
    def hash_bucket_count(self, count: int) -> None:
        self["hash_bucket_count"] = count

    @property
    def hash_group_count(self) -> int:
        return self["hash_group_count"]

    @property
    def ray_custom_resources(self) -> Dict:
        return self["ray_custom_resources"]

    @ray_custom_resources.setter
    def ray_custom_resources(self, res) -> None:
        self["ray_custom_resources"] = res

    @hash_group_count.setter
    def hash_group_count(self, count: int) -> None:
        self["hash_group_count"] = count

    @property
    def primary_keys(self) -> Optional[List[str]]:
        return self.get("primary_keys")

    @primary_keys.setter
    def primary_keys(self, keys: List[str]) -> None:
        self["primary_keys"] = keys

    @property
    def rebase_source_partition_locator(self) -> Optional[PartitionLocator]:
        val = self.get("rebase_source_partition_locator")

        if val and not isinstance(val, PartitionLocator):
            val = PartitionLocator(val)

        return val

    @rebase_source_partition_locator.setter
    def rebase_source_partition_locator(self, locator: PartitionLocator) -> None:
        self["rebase_source_partition_locator"] = locator

    @property
    def rebase_source_partition_high_watermark(self) -> Optional[int]:
        return self.get("rebase_source_partition_high_watermark")

    @rebase_source_partition_high_watermark.setter
    def rebase_source_partition_high_watermark(self, high_watermark: int) -> None:
        self["rebase_source_partition_high_watermark"] = high_watermark

    @property
    def pg_config(self) -> Optional[PlacementGroupConfig]:
        return self.get("pg_config")

    @pg_config.setter
    def pg_config(self, config: PlacementGroupConfig) -> None:
        self["pg_config"] = config

    @property
    def read_kwargs_provider(self) -> Optional[ReadKwargsProvider]:
        return self.get("read_kwargs_provider")

    @read_kwargs_provider.setter
    def read_kwargs_provider(self, kwargs_provider: ReadKwargsProvider) -> None:
        self["read_kwargs_provider"] = kwargs_provider

    @property
    def sort_keys(self) -> Optional[List[SortKey]]:
        return self.get("sort_keys")

    @sort_keys.setter
    def sort_keys(self, keys: List[SortKey]) -> None:
        self["sort_keys"] = keys

    @property
    def metrics_config(self) -> Optional[MetricsConfig]:
        return self.get("metrics_config")

    @metrics_config.setter
    def metrics_config(self, config: MetricsConfig) -> None:
        self["metrics_config"] = config

    @staticmethod
    def json_handler_for_compact_partition_params(obj):
        """
        A handler for the `json.dumps()` function that can be used to serialize sets to JSON.
        If the `set_default()` handler is passed as the `default` argument to the `json.dumps()` function, it will be called whenever a set object is encountered.
        The `set_default()` handler will then serialize the set as a list.
        """
        try:
            if isinstance(obj, set):
                return list(obj)
            elif hasattr(obj, "toJSON"):
                return obj.toJSON()
            else:
                return obj.__dict__
        except Exception:
            return obj.__class__.__name__

    def serialize(self) -> str:
        """
        Serializes itself to a json-formatted string

        Returns:
            The serialized object.

        """
        to_serialize: Dict[str, Any] = {}
        # individually try deepcopy the values from the self dictionary and just use the class name for the value when it is not possible to deepcopy
        for attr, value in self.items():
            try:
                to_serialize[attr] = copy.deepcopy(value)
            except Exception:  # if unable to deep copy the objects like module objects for example then just provide the class name at minimum
                to_serialize[attr] = value.__class__.__name__
        serialized_arguments_compact_partition_args: str = json.dumps(
            to_serialize,
            default=CompactPartitionParams.json_handler_for_compact_partition_params,
        )
        return serialized_arguments_compact_partition_args



from typing import NamedTuple

import numpy as np


class HashBucketResult(NamedTuple):
    hash_bucket_group_to_obj_id_tuple: np.ndarray
    hb_size_bytes: np.int64
    hb_record_count: np.int64
    peak_memory_usage_bytes: np.double
    telemetry_time_in_seconds: np.double
    task_completed_at: np.double



from __future__ import annotations

import pytest


# Benchmarks for retrieving a single column in the Parquet file
SINGLE_COLUMN_BENCHMARKS = {
    "mvp": ("s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet", ["a"]),
    "TPCH-lineitems-200MB-2RG": (
        "s3://daft-public-data/test_fixtures/parquet-dev/daft_200MB_lineitem_chunk.RG-2.parquet",
        ["L_ORDERKEY"],
    ),
}

# Benchmarks for retrieving all columns in the Parquet file
ALL_COLUMN_BENCHMARKS = {
    "mvp": ("s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet", None),
    "TPCH-lineitems-200MB-2RG": (
        "s3://daft-public-data/test_fixtures/parquet-dev/daft_200MB_lineitem_chunk.RG-2.parquet",
        None,
    ),
}


@pytest.mark.benchmark(group="num_rowgroups_single_column")
@pytest.mark.parametrize(
    ["name", "path", "columns"],
    [
        (name, path, columns)
        for name, (path, columns) in SINGLE_COLUMN_BENCHMARKS.items()
    ],
    ids=[name for name in SINGLE_COLUMN_BENCHMARKS],
)
def test_read_parquet_num_rowgroups_single_column(
    name, path, columns, read_fn, benchmark
):
    data = benchmark(read_fn, path, columns=columns)
    if columns is not None:
        assert data.column_names == columns


@pytest.mark.benchmark(group="num_rowgroups_all_columns")
@pytest.mark.parametrize(
    ["name", "path", "columns"],
    [(name, path, columns) for name, (path, columns) in ALL_COLUMN_BENCHMARKS.items()],
    ids=[name for name in ALL_COLUMN_BENCHMARKS],
)
def test_read_parquet_num_rowgroups_all_columns(
    name, path, columns, read_fn, benchmark
):
    data = benchmark(read_fn, path, columns=columns)
    if columns is not None:
        assert data.column_names == columns



import logging
from typing import List, Optional, Iterable

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import uuid
import hashlib
from deltacat.compute.compactor_v2.constants import (
    TOTAL_BYTES_IN_SHA1_HASH,
    PK_DELIMITER,
    MAX_SIZE_OF_RECORD_BATCH_IN_GIB,
)
import time
from deltacat.compute.compactor.model.delta_file_envelope import DeltaFileEnvelope
from deltacat import logs
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.io.object_store import IObjectStore
from deltacat.utils.performance import timed_invocation

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _append_sha1_hash_to_table(table: pa.Table, hash_column: pa.Array) -> pa.Table:
    hash_column_np = hash_column.to_numpy()

    result = []
    for hash_value in hash_column_np:
        assert hash_value is not None, f"Expected non-null primary key"
        result.append(hashlib.sha1(hash_value.encode("utf-8")).hexdigest())

    return sc.append_pk_hash_string_column(table, result)


def _is_sha1_desired(hash_columns: List[pa.Array]) -> bool:
    total_size = 0
    total_len = 0

    for hash_column in hash_columns:
        total_size += hash_column.nbytes
        total_len += len(hash_column)

    logger.info(
        f"Found total length of hash column={total_len} and total_size={total_size}"
    )

    return total_size > TOTAL_BYTES_IN_SHA1_HASH * total_len


def _append_table_by_hash_bucket(
    pki_table: pa.Table, hash_bucket_to_table: np.ndarray
) -> int:

    hb_pk_table, sort_latency = timed_invocation(
        lambda: pki_table.sort_by(sc._HASH_BUCKET_IDX_COLUMN_NAME)
    )
    logger.info(f"Sorting a pk table of length {len(pki_table)} took {sort_latency}s")

    hb_pk_grouped_by, groupby_latency = timed_invocation(
        lambda: hb_pk_table.group_by(sc._HASH_BUCKET_IDX_COLUMN_NAME).aggregate(
            [(sc._HASH_BUCKET_IDX_COLUMN_NAME, "count")]
        )
    )

    logger.info(
        f"Grouping a pki table of length {len(pki_table)} took {groupby_latency}s"
    )

    group_count_array = hb_pk_grouped_by[f"{sc._HASH_BUCKET_IDX_COLUMN_NAME}_count"]
    hb_group_array = hb_pk_grouped_by[sc._HASH_BUCKET_IDX_COLUMN_NAME]

    result_len = 0
    for i, group_count in enumerate(group_count_array):
        hb_idx = hb_group_array[i].as_py()
        pyarrow_table = hb_pk_table.slice(offset=result_len, length=group_count.as_py())
        pyarrow_table = pyarrow_table.drop(
            [sc._HASH_BUCKET_IDX_COLUMN_NAME, sc._PK_HASH_STRING_COLUMN_NAME]
        )
        if hash_bucket_to_table[hb_idx] is None:
            hash_bucket_to_table[hb_idx] = []
        hash_bucket_to_table[hb_idx].append(pyarrow_table)
        result_len += len(pyarrow_table)

    return result_len


def _optimized_group_record_batches_by_hash_bucket(
    pki_table: pa.Table, num_buckets: int
):

    input_table_len = len(pki_table)

    hash_bucket_to_tables = np.empty([num_buckets], dtype="object")
    hb_to_table = np.empty([num_buckets], dtype="object")

    # This split will ensure that the sort is not performed on a very huge table
    # resulting in ArrowInvalid: offset overflow while concatenating arrays
    # Known issue with Arrow: https://github.com/apache/arrow/issues/25822
    table_batches, to_batches_latency = timed_invocation(lambda: pki_table.to_batches())

    logger.info(f"to_batches took {to_batches_latency} for {len(pki_table)} rows")

    current_bytes = 0
    record_batches = []
    result_len = 0
    for record_batch in table_batches:
        current_bytes += record_batch.nbytes
        record_batches.append(record_batch)
        if current_bytes >= MAX_SIZE_OF_RECORD_BATCH_IN_GIB:
            logger.info(
                f"Total number of record batches without exceeding {MAX_SIZE_OF_RECORD_BATCH_IN_GIB} "
                f"is {len(record_batches)} and size {current_bytes}"
            )
            appended_len, append_latency = timed_invocation(
                _append_table_by_hash_bucket,
                pa.Table.from_batches(record_batches),
                hash_bucket_to_tables,
            )
            logger.info(
                f"Appended the hash bucketed batch of {appended_len} in {append_latency}s"
            )

            result_len += appended_len
            current_bytes = 0
            record_batches.clear()

    if record_batches:
        appended_len, append_latency = timed_invocation(
            _append_table_by_hash_bucket,
            pa.Table.from_batches(record_batches),
            hash_bucket_to_tables,
        )
        result_len += appended_len
        current_bytes = 0
        record_batches.clear()

    concat_start = time.monotonic()
    for hb, tables in enumerate(hash_bucket_to_tables):
        if tables:
            assert hb_to_table[hb] is None, f"The HB index is repeated {hb}"
            hb_to_table[hb] = pa.concat_tables(tables)

    concat_end = time.monotonic()
    logger.info(
        f"Total time taken to concat all record batches with length "
        f"{input_table_len}: {concat_end - concat_start}s"
    )

    assert (
        input_table_len == result_len
    ), f"Grouping has resulted in record loss as {result_len} != {input_table_len}"

    return hb_to_table


def group_by_pk_hash_bucket(
    table: pa.Table, num_buckets: int, primary_keys: List[str]
) -> np.ndarray:
    table = generate_pk_hash_column([table], primary_keys, requires_sha1=True)[0]

    # group hash bucket record indices
    result = group_record_indices_by_hash_bucket(
        table,
        num_buckets,
    )

    return result


def generate_pk_hash_column(
    tables: List[pa.Table],
    primary_keys: Optional[List[str]] = None,
    requires_sha1: bool = False,
) -> List[pa.Table]:
    """
    Returns a new table list after generating the primary key hash if desired.

    1. If there are no primary keys, each hash will be unique uuid/sha1 hex
    2. If there are more than 0 primary keys, returns a table with pk hash column appended.
    """

    def _generate_pk_hash(table: pa.Table) -> pa.Array:
        pk_columns = []
        for pk_name in primary_keys:
            pk_columns.append(pc.cast(table[pk_name], pa.string()))

        pk_columns.append(PK_DELIMITER)
        hash_column = pc.binary_join_element_wise(*pk_columns)
        return hash_column

    def _generate_uuid(table: pa.Table) -> pa.Array:
        hash_column = pa.array(
            [uuid.uuid4().hex for _ in range(len(table))], pa.string()
        )
        return hash_column

    start = time.monotonic()

    hash_column_list = []

    can_sha1 = False
    if primary_keys:
        hash_column_list = [_generate_pk_hash(table) for table in tables]

        can_sha1 = requires_sha1 or _is_sha1_desired(hash_column_list)
    else:
        hash_column_list = [_generate_uuid(table) for table in tables]

    logger.info(
        f"can_generate_sha1={can_sha1} for the table and requires_sha1={requires_sha1}"
    )

    result = []

    total_len = 0
    total_size = 0
    for index, table in enumerate(tables):
        if can_sha1:
            table = _append_sha1_hash_to_table(table, hash_column_list[index])
        else:
            table = table.append_column(
                sc._PK_HASH_STRING_COLUMN_FIELD, hash_column_list[index]
            )

        total_len += len(table)
        total_size += hash_column_list[index].nbytes

        result.append(table)

    end = time.monotonic()

    logger.info(
        f"Took {end - start}s to generate pk hash of len: {total_len}"
        f" for size: {total_size} bytes"
    )

    return result


def group_record_indices_by_hash_bucket(
    pki_table: pa.Table, num_buckets: int
) -> np.ndarray:
    """
    Groups the record indices by it's corresponding hash bucket. Hence, this method may
    create num_buckets tables as a result.
    """

    input_table_len = len(pki_table)

    hash_bucket_id_col_list = np.empty([input_table_len], dtype="int32")
    bucketing_start_time = time.monotonic()

    for index, hash_value in enumerate(sc.pk_hash_string_column_np(pki_table)):
        hash_bucket = pk_digest_to_hash_bucket_index(hash_value, num_buckets)
        hash_bucket_id_col_list[index] = hash_bucket

    pki_table = sc.append_hash_bucket_idx_col(pki_table, hash_bucket_id_col_list)
    bucketing_end_time = time.monotonic()

    logger.info(
        f"Took {bucketing_end_time - bucketing_start_time}s to generate the "
        f"hb index for {len(pki_table)} rows"
    )

    result, group_latency = timed_invocation(
        _optimized_group_record_batches_by_hash_bucket,
        pki_table=pki_table,
        num_buckets=num_buckets,
    )

    logger.info(
        f"Final grouping of table with {input_table_len} records took: {group_latency}s"
    )

    return result


def group_hash_bucket_indices(
    hash_bucket_object_groups: np.ndarray,
    num_buckets: int,
    num_groups: int,
    object_store: Optional[IObjectStore] = None,
) -> np.ndarray:
    """
    This method persists all tables for a given hash bucket into the object store
    and returns the object references for each hash group.
    """

    hash_bucket_group_to_obj_id_size_tuple = np.empty([num_groups], dtype="object")

    if hash_bucket_object_groups is None:
        return hash_bucket_group_to_obj_id_size_tuple

    hb_group_to_object = np.empty([num_groups], dtype="object")
    hash_group_to_size = np.empty([num_groups], dtype="int64")
    hash_group_to_num_rows = np.empty([num_groups], dtype="int64")

    for hb_index, obj in enumerate(hash_bucket_object_groups):
        if obj:
            hb_group = hash_bucket_index_to_hash_group_index(hb_index, num_groups)
            if hb_group_to_object[hb_group] is None:
                hb_group_to_object[hb_group] = np.empty([num_buckets], dtype="object")
                hash_group_to_size[hb_group] = np.int64(0)
                hash_group_to_num_rows[hb_group] = np.int64(0)
            hb_group_to_object[hb_group][hb_index] = obj
            for dfe in obj:
                casted_dfe: DeltaFileEnvelope = dfe
                hash_group_to_size[hb_group] += casted_dfe.table_size_bytes
                hash_group_to_num_rows[hb_group] += casted_dfe.table_num_rows

    for hb_group, obj in enumerate(hb_group_to_object):
        if obj is None:
            continue
        object_ref = object_store.put(obj)
        hash_bucket_group_to_obj_id_size_tuple[hb_group] = (
            object_ref,
            hash_group_to_size[hb_group],
            hash_group_to_num_rows[hb_group],
        )
        del object_ref

    _, close_latency = timed_invocation(object_store.close)
    logger.info(f"Active connections to the object store closed in {close_latency}")

    return hash_bucket_group_to_obj_id_size_tuple


def hash_bucket_index_to_hash_group_index(hb_index: int, num_groups: int) -> int:
    return hb_index % num_groups


def hash_group_index_to_hash_bucket_indices(
    hb_group: int, num_buckets: int, num_groups: int
) -> Iterable[int]:

    if hb_group > num_buckets:
        return []

    return range(hb_group, num_buckets, num_groups)


def pk_digest_to_hash_bucket_index(digest: str, num_buckets: int) -> int:
    """
    Generates the hash bucket index from the given digest.
    """
    return int(digest, 16) % num_buckets



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, List, Optional

from deltacat.storage.model.locator import Locator
from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.storage.model.table import TableLocator
from deltacat.storage.model.table_version import TableVersionLocator
from deltacat.storage.model.types import CommitState


class Stream(dict):
    @staticmethod
    def of(
        locator: Optional[StreamLocator],
        partition_keys: Optional[List[Dict[str, Any]]],
        state: Optional[CommitState] = None,
        previous_stream_digest: Optional[bytes] = None,
    ) -> Stream:
        stream = Stream()
        stream.locator = locator
        stream.partition_keys = partition_keys
        stream.state = state
        stream.previous_stream_digest = previous_stream_digest
        return stream

    @property
    def locator(self) -> Optional[StreamLocator]:
        val: Dict[str, Any] = self.get("streamLocator")
        if val is not None and not isinstance(val, StreamLocator):
            self.locator = val = StreamLocator(val)
        return val

    @locator.setter
    def locator(self, stream_locator: Optional[StreamLocator]) -> None:
        self["streamLocator"] = stream_locator

    @property
    def partition_keys(self) -> Optional[List[Dict[str, Any]]]:
        return self.get("partitionKeys")

    @partition_keys.setter
    def partition_keys(self, partition_keys: Optional[List[Dict[str, Any]]]) -> None:
        self["partitionKeys"] = partition_keys

    @property
    def previous_stream_digest(self) -> Optional[str]:
        return self.get("previousStreamDigest")

    @previous_stream_digest.setter
    def previous_stream_digest(self, previous_stream_digest: Optional[str]) -> None:
        self["previousStreamDigest"] = previous_stream_digest

    @property
    def state(self) -> Optional[CommitState]:
        state = self.get("state")
        return None if state is None else CommitState(state)

    @state.setter
    def state(self, state: Optional[CommitState]) -> None:
        self["state"] = state

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        stream_locator = self.locator
        if stream_locator:
            return stream_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        stream_locator = self.locator
        if stream_locator:
            return stream_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        stream_locator = self.locator
        if stream_locator:
            return stream_locator.table_version_locator
        return None

    @property
    def stream_id(self) -> Optional[str]:
        stream_locator = self.locator
        if stream_locator:
            return stream_locator.stream_id
        return None

    @property
    def namespace(self) -> Optional[str]:
        stream_locator = self.locator
        if stream_locator:
            return stream_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        stream_locator = self.locator
        if stream_locator:
            return stream_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        stream_locator = self.locator
        if stream_locator:
            return stream_locator.table_version
        return None

    def validate_partition_values(self, partition_values: Optional[List[Any]]):
        # TODO (pdames): ensure value data types match key data types
        partition_keys = self.partition_keys
        num_keys = len(partition_keys) if partition_keys else 0
        num_values = len(partition_values) if partition_values else 0
        if num_values != num_keys:
            raise ValueError(
                f"Found {num_values} partition values but "
                f"{num_keys} partition keys: {self}"
            )


class StreamLocator(Locator, dict):
    @staticmethod
    def of(
        table_version_locator: Optional[TableVersionLocator],
        stream_id: Optional[str],
        storage_type: Optional[str],
    ) -> StreamLocator:
        """
        Creates a table version Stream Locator. All input parameters are
        case-sensitive.
        """
        stream_locator = StreamLocator()
        stream_locator.table_version_locator = table_version_locator
        stream_locator.stream_id = stream_id
        stream_locator.storage_type = storage_type
        return stream_locator

    @staticmethod
    def at(
        namespace: Optional[str],
        table_name: Optional[str],
        table_version: Optional[str],
        stream_id: Optional[str],
        storage_type: Optional[str],
    ) -> StreamLocator:
        table_version_locator = TableVersionLocator.at(
            namespace,
            table_name,
            table_version,
        )
        return StreamLocator.of(
            table_version_locator,
            stream_id,
            storage_type,
        )

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        val: Dict[str, Any] = self.get("tableVersionLocator")
        if val is not None and not isinstance(val, TableVersionLocator):
            self.table_version_locator = val = TableVersionLocator(val)
        return val

    @table_version_locator.setter
    def table_version_locator(
        self, table_version_locator: Optional[TableVersionLocator]
    ) -> None:
        self["tableVersionLocator"] = table_version_locator

    @property
    def stream_id(self) -> Optional[str]:
        return self.get("streamId")

    @stream_id.setter
    def stream_id(self, stream_id: Optional[str]) -> None:
        self["streamId"] = stream_id

    @property
    def storage_type(self) -> Optional[str]:
        return self.get("storageType")

    @storage_type.setter
    def storage_type(self, storage_type: Optional[str]) -> None:
        self["storageType"] = storage_type

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        table_version_locator = self.table_version_locator
        if table_version_locator:
            return table_version_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        table_version_locator = self.table_version_locator
        if table_version_locator:
            return table_version_locator.table_locator
        return None

    @property
    def namespace(self) -> Optional[str]:
        table_version_locator = self.table_version_locator
        if table_version_locator:
            return table_version_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        table_version_locator = self.table_version_locator
        if table_version_locator:
            return table_version_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        table_version_locator = self.table_version_locator
        if table_version_locator:
            return table_version_locator.table_version
        return None

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        tvl_hexdigest = self.table_version_locator.hexdigest()
        stream_id = self.stream_id
        storage_type = self.storage_type
        return f"{tvl_hexdigest}|{stream_id}|{storage_type}"



import io
from collections import deque
from typing import List, Any
from ray import cloudpickle


def dump_into_chunks(obj: Any, max_size_bytes: int) -> List[bytes]:
    """
    This is an optimized version of dump which only requires additional
    BUFFER_SIZE amount of memory which can be configured.
    """
    assert max_size_bytes > 0, "max_size_bytes must be greater than zero"

    bytes_io = io.BytesIO()
    cloudpickle.dump(obj, bytes_io)
    total_size = bytes_io.tell()

    if total_size <= max_size_bytes:
        return [bytes_io.getvalue()]

    bytes_io.seek(0, 2)  # seek to end
    result = deque()
    observed_total_size = 0
    current_size = bytes_io.tell()

    while current_size > 0:
        to_read_size = min(max_size_bytes, current_size)
        bytes_io.seek(-to_read_size, 1)
        cur = bytes_io.read(to_read_size)
        observed_total_size += len(cur)
        bytes_io.seek(-to_read_size, 1)
        bytes_io.truncate()
        result.appendleft(cur)
        current_size = bytes_io.tell()

    assert (
        observed_total_size == total_size
    ), f"Not all bytes were split as {observed_total_size} != {total_size}"

    return list(result)



from unittest import TestCase
from ray import cloudpickle
from deltacat.utils.cloudpickle import dump_into_chunks
import pyarrow as pa
import numpy as np


class TestDumpIntoChunks(TestCase):
    def test_dump_into_chunks_sanity(self):
        table = self._get_test_table()
        chunks = dump_into_chunks(
            table, table.nbytes // 100
        )  # divides into atleast 100 chunks
        serialized = cloudpickle.dumps(table)
        concatenated_serialized = self._concat_bytes_chunks(chunks)

        self.assertEqual(serialized, concatenated_serialized)
        self.assertGreater(len(chunks), 100)

    def test_dump_into_chunks_when_max_size_zero(self):
        table = self._get_test_table()
        self.assertRaises(AssertionError, lambda: dump_into_chunks(table, 0))

    def test_dump_into_chunks_when_max_size_greater_or_equal_table_size(self):
        table = self._get_test_table()
        chunks = dump_into_chunks(table, table.nbytes)
        serialized = cloudpickle.dumps(table)
        concatenated_serialized = self._concat_bytes_chunks(chunks)

        self.assertEqual(serialized, concatenated_serialized)
        self.assertLess(len(chunks), 3)

    def _get_test_table(self):
        col1 = np.random.randint(100000, size=100000)
        return pa.table({"pk_hash": col1, "some_col": col1, "some_col2": col1})

    def _concat_bytes_chunks(self, chunks):
        result = bytearray()
        for chunk in chunks:
            result.extend(chunk)

        return result



import logging
from typing import List, Optional, Tuple

import numpy as np
import pyarrow as pa
from ray.types import ObjectRef

from deltacat import logs
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.io.object_store import IObjectStore

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

# TODO: Deprecate this module in the favor of compactor_v2


def group_record_indices_by_hash_bucket(
    pki_table: pa.Table, num_buckets: int
) -> np.ndarray:

    hash_bucket_to_indices = np.empty([num_buckets], dtype="object")
    record_index = 0
    for digest in sc.pk_hash_column_np(pki_table):
        hash_bucket = pk_digest_to_hash_bucket_index(digest, num_buckets)
        if hash_bucket_to_indices[hash_bucket] is None:
            hash_bucket_to_indices[hash_bucket] = []
        hash_bucket_to_indices[hash_bucket].append(record_index)
        record_index += 1
    return hash_bucket_to_indices


def group_hash_bucket_indices(
    hash_bucket_object_groups: np.ndarray,
    num_buckets: int,
    num_groups: int,
    object_store: Optional[IObjectStore] = None,
) -> Tuple[np.ndarray, List[ObjectRef]]:
    """
    Groups all the ObjectRef that belongs to a particular hash bucket group and hash bucket index.
    """

    object_refs = []
    hash_bucket_group_to_obj_id = np.empty([num_groups], dtype="object")

    if hash_bucket_object_groups is None:
        return hash_bucket_group_to_obj_id, object_refs

    hb_group_to_object = np.empty([num_groups], dtype="object")
    for hb_index, obj in enumerate(hash_bucket_object_groups):
        if obj:
            hb_group = hb_index % num_groups
            if hb_group_to_object[hb_group] is None:
                hb_group_to_object[hb_group] = np.empty([num_buckets], dtype="object")
            hb_group_to_object[hb_group][hb_index] = obj

    for hb_group, obj in enumerate(hb_group_to_object):
        if obj is None:
            continue
        object_ref = object_store.put(obj)
        object_refs.append(object_ref)
        hash_bucket_group_to_obj_id[hb_group] = object_ref
        del object_ref
    return hash_bucket_group_to_obj_id, object_refs


def pk_digest_to_hash_bucket_index(digest, num_buckets: int) -> int:
    """
    Deterministically get the hash bucket a particular digest belongs to
    based on number of total hash buckets.
    """

    return int.from_bytes(digest, "big") % num_buckets



import json
import logging
from typing import List

from deltacat import logs
from deltacat.aws import s3u as s3_utils
from deltacat.compute.stats.models.delta_column_stats import DeltaColumnStats
from deltacat.compute.stats.models.manifest_entry_stats import ManifestEntryStats
from deltacat.storage import DeltaLocator
from deltacat.utils.common import sha1_hexdigest

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def get_manifest_stats_s3_url(
    bucket: str, column_name: str, delta_locator: DeltaLocator
) -> str:
    """Returns the S3 URL path to the column-oriented delta stats

    Args:
        bucket: The S3 bucket
        column_name: The name of the column to look up stats for
        delta_locator: The reference to the delta corresponding to the manifest entries

    Returns:
        A S3 URL path
    """
    stats_column_id = f"{delta_locator.canonical_string()}|{column_name}"
    stats_column_hexdigest = sha1_hexdigest(stats_column_id.encode("utf-8"))
    base_path = s3_utils.parse_s3_url(bucket).url
    return f"{base_path}/{stats_column_hexdigest}.json"


def read_manifest_stats_by_columns(
    bucket: str, column_names: List[str], delta_locator: DeltaLocator
) -> List[DeltaColumnStats]:
    """Fetch a list of delta column stats by reading each column-oriented delta stats file from S3

    Args:
        bucket: The S3 bucket
        column_names: A list of column names to look up stats for
        delta_locator: The reference to the delta corresponding to the manifest entries

    Returns:
        A list of delta column stats
    """
    return [
        DeltaColumnStats.of(
            column, read_manifest_stats_file(bucket, column, delta_locator)
        )
        for column in column_names
    ]


def read_manifest_stats_file(
    bucket: str, column_name: str, delta_locator: DeltaLocator
) -> ManifestEntryStats:
    """Read a manifest entry stats from S3

    Args:
        bucket: The S3 bucket
        column_name: The name of the column to look up stats for
        delta_locator: The reference to the delta corresponding to the manifest entries

    Returns:
        A container that holds a list of manifest entry stats for the given column name
    """

    stats_completion_file_url = get_manifest_stats_s3_url(
        bucket, column_name, delta_locator
    )
    logger.info(f"reading stats completion file from: {stats_completion_file_url}")
    stats_completion_info_file = None
    result = s3_utils.download(stats_completion_file_url, fail_if_not_found=False)
    if result:
        json_str = result["Body"].read().decode("utf-8")
        stats_completion_info_file = ManifestEntryStats(json.loads(json_str))
        logger.info(f"read stats completion info: {stats_completion_info_file}")
    return stats_completion_info_file


def write_manifest_stats_file(
    bucket: str, column_name: str, manifest_entry_stats: ManifestEntryStats
) -> None:
    """Write a manifest entry stats into S3

    Args:
        bucket: The S3 bucket
        column_name: The name of the column which represents this manifest entry stats
        manifest_entry_stats: The manifest entry stats to serialize and store into S3
    """
    logger.info(f"writing stats completion file contents: {manifest_entry_stats}")
    stats_completion_file_s3_url = get_manifest_stats_s3_url(
        bucket,
        column_name,
        manifest_entry_stats.delta_locator,
    )
    logger.info(f"writing stats completion file to: {stats_completion_file_s3_url}")
    s3_utils.upload(stats_completion_file_s3_url, str(json.dumps(manifest_entry_stats)))
    logger.info(f"stats completion file written to: {stats_completion_file_s3_url}")



from typing import List

from deltacat.utils.common import env_integer, env_string

DAFT_MAX_S3_CONNECTIONS_PER_FILE = env_integer("DAFT_MAX_S3_CONNECTIONS_PER_FILE", 8)
BOTO_MAX_RETRIES = env_integer("BOTO_MAX_RETRIES", 15)
TIMEOUT_ERROR_CODES: List[str] = ["ReadTimeoutError", "ConnectTimeoutError"]
AWS_REGION = env_string("AWS_REGION", "us-east-1")



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import numpy as np
import pyarrow as pa

from deltacat.storage import DeltaType, LocalTable

from typing import Optional

DeltaFileEnvelopeGroups = np.ndarray


class DeltaFileEnvelope(dict):
    @staticmethod
    def of(
        stream_position: int,
        delta_type: DeltaType,
        table: LocalTable,
        file_index: int = None,
        is_src_delta: np.bool_ = True,
        file_record_count: Optional[int] = None,
    ) -> DeltaFileEnvelope:
        """Static factory builder for a Delta File Envelope
        `
        Args:
            stream_position: Stream position of a delta.
            file_index: Manifest file index number of a delta.
            delta_type: A delta type.
            table: The table object that represents the delta file.
            is_src_delta: True if this Delta File Locator is
                pointing to a file from the uncompacted source table, False if
                this Locator is pointing to a file in the compacted destination
                table.
        Returns:
            A delta file envelope.

        """
        if stream_position is None:
            raise ValueError("Missing delta file envelope stream position.")
        if delta_type is None:
            raise ValueError("Missing Delta file envelope delta type.")
        if table is None:
            raise ValueError("Missing Delta file envelope table.")
        delta_file_envelope = DeltaFileEnvelope()
        delta_file_envelope["streamPosition"] = stream_position
        delta_file_envelope["fileIndex"] = file_index
        delta_file_envelope["deltaType"] = delta_type.value
        delta_file_envelope["table"] = table
        delta_file_envelope["is_src_delta"] = is_src_delta
        delta_file_envelope["file_record_count"] = file_record_count
        return delta_file_envelope

    @property
    def stream_position(self) -> int:
        return self["streamPosition"]

    @property
    def file_index(self) -> int:
        return self["fileIndex"]

    @property
    def delta_type(self) -> DeltaType:
        return DeltaType(self["deltaType"])

    @property
    def table(self) -> LocalTable:
        return self["table"]

    @property
    def is_src_delta(self) -> np.bool_:
        return self["is_src_delta"]

    @property
    def file_record_count(self) -> int:
        return self["file_record_count"]

    @property
    def table_size_bytes(self) -> int:
        if isinstance(self.table, pa.Table):
            return self.table.nbytes
        else:
            raise ValueError(
                f"Table type: {type(self.table)} not for supported for size method."
            )

    @property
    def table_num_rows(self) -> int:
        return len(self.table)



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import List


class PyArrowWriteResult(dict):
    @staticmethod
    def of(
        file_count: int, pyarrow_bytes: int, file_bytes: int, record_count: int
    ) -> PyArrowWriteResult:
        pawr = PyArrowWriteResult()
        pawr["files"] = file_count
        pawr["paBytes"] = pyarrow_bytes
        pawr["fileBytes"] = file_bytes
        pawr["records"] = record_count
        return pawr

    @staticmethod
    def union(results: List[PyArrowWriteResult]) -> PyArrowWriteResult:
        """
        Create a new PyArrowWriteResult containing all results from all input
        PyArrowWriteResults.
        """
        return PyArrowWriteResult.of(
            sum([result.files for result in results]),
            sum([result.pyarrow_bytes for result in results]),
            sum([result.file_bytes for result in results]),
            sum([result.records for result in results]),
        )

    @property
    def files(self) -> int:
        return self["files"]

    @property
    def pyarrow_bytes(self) -> int:
        return self["paBytes"]

    @property
    def file_bytes(self) -> int:
        return self["fileBytes"]

    @property
    def records(self) -> int:
        return self["records"]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import bz2
import gzip
import io
import logging
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional
from pyarrow.parquet import ParquetFile
from deltacat.exceptions import ValidationError

import pyarrow as pa
from fsspec import AbstractFileSystem
from pyarrow import csv as pacsv
from pyarrow import feather as paf
from pyarrow import json as pajson
from pyarrow import parquet as papq
from ray.data.datasource import BlockWritePathProvider
from deltacat.utils.s3fs import create_s3_file_system

from deltacat import logs
from deltacat.types.media import (
    DELIMITED_TEXT_CONTENT_TYPES,
    TABULAR_CONTENT_TYPES,
    ContentEncoding,
    ContentType,
)
from deltacat.types.partial_download import (
    PartialFileDownloadParams,
    PartialParquetParameters,
)
from deltacat.utils.common import ContentTypeKwargsProvider, ReadKwargsProvider
from deltacat.utils.performance import timed_invocation
from deltacat.utils.daft import daft_s3_file_to_table
from deltacat.utils.schema import coerce_pyarrow_table_to_schema
from deltacat.utils.arguments import (
    sanitize_kwargs_to_callable,
    sanitize_kwargs_by_supported_kwargs,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

RAISE_ON_EMPTY_CSV_KWARG = "raise_on_empty_csv"
READER_TYPE_KWARG = "reader_type"


def _filter_schema_for_columns(schema: pa.Schema, columns: List[str]) -> pa.Schema:

    target_schema_fields = []

    for column_name in columns:
        index = schema.get_field_index(column_name)

        if index != -1:
            target_field = schema.field(index)
            target_schema_fields.append(target_field)

    target_schema = pa.schema(target_schema_fields, metadata=schema.metadata)

    return target_schema


def pyarrow_read_csv(*args, **kwargs) -> pa.Table:
    try:
        new_kwargs = sanitize_kwargs_by_supported_kwargs(
            ["read_options", "parse_options", "convert_options", "memory_pool"], kwargs
        )
        return pacsv.read_csv(*args, **new_kwargs)
    except pa.lib.ArrowInvalid as e:
        if e.__str__() == "Empty CSV file" and not kwargs.get(RAISE_ON_EMPTY_CSV_KWARG):
            schema = None
            if (
                "convert_options" in kwargs
                and kwargs["convert_options"].column_types is not None
            ):
                schema = kwargs["convert_options"].column_types
                if not isinstance(schema, pa.Schema):
                    schema = pa.schema(schema)
                if kwargs["convert_options"].include_columns:
                    schema = _filter_schema_for_columns(
                        schema, kwargs["convert_options"].include_columns
                    )
                elif (
                    kwargs.get("read_options") is not None
                    and kwargs["read_options"].column_names
                ):
                    schema = _filter_schema_for_columns(
                        schema, kwargs["read_options"].column_names
                    )

            else:
                logger.debug(
                    "Schema not specified in the kwargs."
                    " Hence, schema could not be inferred from the empty CSV."
                )

            logger.debug(f"Read CSV empty schema being used: {schema}")
            return pa.Table.from_pylist([], schema=schema)
        raise e


CONTENT_TYPE_TO_PA_READ_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: pyarrow_read_csv,
    ContentType.TSV.value: pyarrow_read_csv,
    ContentType.CSV.value: pyarrow_read_csv,
    ContentType.PSV.value: pyarrow_read_csv,
    ContentType.PARQUET.value: papq.read_table,
    ContentType.FEATHER.value: paf.read_table,
    # Pyarrow.orc is disabled in Pyarrow 0.15, 0.16:
    # https://issues.apache.org/jira/browse/ARROW-7811
    # ContentType.ORC.value: paorc.ContentType.ORCFile,
    ContentType.JSON.value: pajson.read_json,
}


def write_feather(
    table: pa.Table, path: str, *, filesystem: AbstractFileSystem, **kwargs
) -> None:

    with filesystem.open(path, "wb") as f:
        paf.write_feather(table, f, **kwargs)


def write_csv(
    table: pa.Table, path: str, *, filesystem: AbstractFileSystem, **kwargs
) -> None:

    with filesystem.open(path, "wb") as f:
        # TODO (pdames): Add support for client-specified compression types.
        with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
            if kwargs.get("write_options") is None:
                # column names are kept in table metadata, so omit header
                kwargs["write_options"] = pacsv.WriteOptions(include_header=False)
            pacsv.write_csv(table, out, **kwargs)


CONTENT_TYPE_TO_PA_WRITE_FUNC: Dict[str, Callable] = {
    # TODO (pdames): add support for other delimited text content types as
    #  pyarrow adds support for custom delimiters, escaping, and None value
    #  representations to pyarrow.csv.WriteOptions.
    ContentType.CSV.value: write_csv,
    ContentType.PARQUET.value: papq.write_table,
    ContentType.FEATHER.value: write_feather,
}


def content_type_to_reader_kwargs(content_type: str) -> Dict[str, Any]:
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "parse_options": pacsv.ParseOptions(delimiter="\t", quote_char=False),
            "convert_options": pacsv.ConvertOptions(
                null_values=[""],  # pyarrow defaults are ["", "NULL", "null"]
                strings_can_be_null=True,
            ),
        }
    if content_type == ContentType.TSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter="\t")}
    if content_type == ContentType.CSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter=",")}
    if content_type == ContentType.PSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter="|")}
    if content_type in {
        ContentType.PARQUET.value,
        ContentType.FEATHER.value,
        ContentType.JSON.value,
    }:
        return {}
    # Pyarrow.orc is disabled in Pyarrow 0.15, 0.16:
    # https://issues.apache.org/jira/browse/ARROW-7811
    # if DataTypes.ContentType.ORC:
    #   return {},
    raise ValueError(f"Unsupported content type: {content_type}")


# TODO (pdames): add deflate and snappy
ENCODING_TO_FILE_INIT: Dict[str, Callable] = {
    ContentEncoding.GZIP.value: partial(gzip.open, mode="rb"),
    ContentEncoding.BZIP2.value: partial(bz2.open, mode="rb"),
    ContentEncoding.IDENTITY.value: lambda s3_file: s3_file,
}


def slice_table(table: pa.Table, max_len: Optional[int]) -> List[pa.Table]:
    """
    Iteratively create 0-copy table slices.
    """
    if max_len is None:
        return [table]
    tables = []
    offset = 0
    records_remaining = len(table)
    while records_remaining > 0:
        records_this_entry = min(max_len, records_remaining)
        tables.append(table.slice(offset, records_this_entry))
        records_remaining -= records_this_entry
        offset += records_this_entry
    return tables


class ReadKwargsProviderPyArrowCsvPureUtf8(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that reads columns of delimited text files
    as UTF-8 strings (i.e. disables type inference). Useful for ensuring
    lossless reads of UTF-8 delimited text datasets and improving read
    performance in cases where type casting is not required."""

    def __init__(self, include_columns: Optional[Iterable[str]] = None):
        self.include_columns = include_columns

    def _get_kwargs(self, content_type: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            convert_options: pacsv.ConvertOptions = kwargs.get("convert_options")
            if convert_options is None:
                convert_options = pacsv.ConvertOptions()
            # read only the included columns as strings?
            column_names = (
                self.include_columns
                if self.include_columns
                else convert_options.include_columns
            )
            if not column_names:
                # read all columns as strings?
                read_options: pacsv.ReadOptions = kwargs.get("read_options")
                if read_options and read_options.column_names:
                    column_names = read_options.column_names
                else:
                    raise ValueError("No column names found!")
            convert_options.column_types = {
                column_name: pa.string() for column_name in column_names
            }
            kwargs["convert_options"] = convert_options
        return kwargs


class ReadKwargsProviderPyArrowSchemaOverride(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that explicitly maps column names to column types when
    loading dataset files into a PyArrow table. Disables the default type inference
    behavior on the defined columns."""

    def __init__(
        self,
        schema: Optional[pa.Schema] = None,
        pq_coerce_int96_timestamp_unit: Optional[str] = None,
        parquet_reader_type: Optional[str] = None,
    ):
        """

        Args:
            schema: The schema to use for reading the dataset.
                If unspecified, the schema will be inferred from the source.
            pq_coerce_int96_timestamp_unit: When reading from parquet files, cast timestamps that are stored in INT96
                format to a particular resolution (e.g. 'ms').

        """
        self.schema = schema
        self.pq_coerce_int96_timestamp_unit = pq_coerce_int96_timestamp_unit
        self.parquet_reader_type = parquet_reader_type

    def _get_kwargs(self, content_type: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            convert_options = kwargs.get("convert_options", pacsv.ConvertOptions())
            if self.schema:
                convert_options.column_types = self.schema
            kwargs["convert_options"] = convert_options
        elif content_type == ContentType.PARQUET:
            # kwargs here are passed into `pyarrow.parquet.read_table`.
            # Only supported in PyArrow 8.0.0+
            if self.schema:
                kwargs["schema"] = self.schema

            # Coerce deprecated int96 timestamp to millisecond if unspecified
            if self.pq_coerce_int96_timestamp_unit is not None:
                kwargs[
                    "coerce_int96_timestamp_unit"
                ] = self.pq_coerce_int96_timestamp_unit

            if self.parquet_reader_type:
                kwargs["reader_type"] = self.parquet_reader_type
            else:
                kwargs["reader_type"] = "daft"

        return kwargs


def _add_column_kwargs(
    content_type: str,
    column_names: Optional[List[str]],
    include_columns: Optional[List[str]],
    kwargs: Dict[str, Any],
):

    if content_type in DELIMITED_TEXT_CONTENT_TYPES:
        read_options: pacsv.ReadOptions = kwargs.get("read_options")
        if read_options is None:
            read_options = pacsv.ReadOptions()
        if column_names:
            read_options.column_names = column_names
        else:
            read_options.autogenerate_column_names = True
        kwargs["read_options"] = read_options
        convert_options: pacsv.ConvertOptions = kwargs.get("convert_options")
        if convert_options is None:
            convert_options = pacsv.ConvertOptions()
        if include_columns:
            convert_options.include_columns = include_columns
        kwargs["convert_options"] = convert_options
    else:
        if content_type in TABULAR_CONTENT_TYPES:
            kwargs["columns"] = include_columns
        else:
            if include_columns:
                logger.warning(
                    f"Ignoring request to include columns {include_columns} "
                    f"for non-tabular content type {content_type}"
                )


def s3_partial_parquet_file_to_table(
    s3_url: str,
    content_type: str,
    content_encoding: str,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialParquetParameters] = None,
    **s3_client_kwargs,
) -> pa.Table:

    assert (
        partial_file_download_params is not None
    ), "Partial parquet params must not be None"
    assert (
        partial_file_download_params.row_groups_to_download is not None
    ), "No row groups to download"

    pq_file = s3_file_to_parquet(
        s3_url=s3_url,
        content_type=content_type,
        content_encoding=content_encoding,
        partial_file_download_params=partial_file_download_params,
        pa_read_func_kwargs_provider=pa_read_func_kwargs_provider,
        **s3_client_kwargs,
    )

    table, latency = timed_invocation(
        pq_file.read_row_groups,
        partial_file_download_params.row_groups_to_download,
        columns=include_columns or column_names,
    )

    logger.debug(f"Successfully read from s3_url={s3_url} in {latency}s")

    kwargs = {}

    if pa_read_func_kwargs_provider:
        kwargs = pa_read_func_kwargs_provider(content_type, kwargs)

    # Note: ordering is consistent with the `input_schema` if provided
    if kwargs.get("schema") is not None:
        input_schema = kwargs.get("schema")
        if include_columns is not None:
            input_schema = pa.schema(
                [input_schema.field(col) for col in include_columns],
                metadata=input_schema.metadata,
            )
        elif column_names is not None:
            input_schema = pa.schema(
                [input_schema.field(col) for col in column_names],
                metadata=input_schema.metadata,
            )
        coerced_table, coerce_latency = timed_invocation(
            coerce_pyarrow_table_to_schema, table, input_schema
        )

        logger.debug(
            f"Coercing the PyArrow table of len {len(coerced_table)} "
            f"into passed schema took {coerce_latency}s"
        )

        return coerced_table

    return table


def s3_file_to_table(
    s3_url: str,
    content_type: str,
    content_encoding: str,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    **s3_client_kwargs,
) -> pa.Table:

    logger.debug(
        f"Reading {s3_url} to PyArrow. Content type: {content_type}. "
        f"Encoding: {content_encoding}"
    )

    kwargs = content_type_to_reader_kwargs(content_type)
    _add_column_kwargs(content_type, column_names, include_columns, kwargs)

    if pa_read_func_kwargs_provider is not None:
        kwargs = pa_read_func_kwargs_provider(content_type, kwargs)

    if (
        content_type == ContentType.PARQUET.value
        and content_encoding == ContentEncoding.IDENTITY.value
    ):
        logger.debug(
            f"Performing read using parquet reader for encoding={content_encoding} "
            f"and content_type={content_type}"
        )

        parquet_reader_func = None
        if kwargs.get(READER_TYPE_KWARG, "daft") == "daft":
            parquet_reader_func = daft_s3_file_to_table
        elif partial_file_download_params and isinstance(
            partial_file_download_params, PartialParquetParameters
        ):
            parquet_reader_func = s3_partial_parquet_file_to_table

        if parquet_reader_func is not None:
            return parquet_reader_func(
                s3_url=s3_url,
                content_type=content_type,
                content_encoding=content_encoding,
                column_names=column_names,
                include_columns=include_columns,
                pa_read_func_kwargs_provider=pa_read_func_kwargs_provider,
                partial_file_download_params=partial_file_download_params,
                **s3_client_kwargs,
            )

        if READER_TYPE_KWARG in kwargs:
            kwargs.pop(READER_TYPE_KWARG)

    filesystem = io
    if s3_url.startswith("s3://"):
        filesystem = create_s3_file_system(s3_client_kwargs)

    logger.debug(f"Read S3 object from {s3_url} using filesystem: {filesystem}")
    input_file_init = ENCODING_TO_FILE_INIT[content_encoding]
    pa_read_func = CONTENT_TYPE_TO_PA_READ_FUNC[content_type]

    with filesystem.open(s3_url, "rb") as s3_file, input_file_init(
        s3_file
    ) as input_file:
        args = [input_file]
        logger.debug(f"Reading {s3_url} via {pa_read_func} with kwargs: {kwargs}")
        table, latency = timed_invocation(pa_read_func, *args, **kwargs)
        logger.debug(f"Time to read {s3_url} into PyArrow table: {latency}s")
        return table


def s3_file_to_parquet(
    s3_url: str,
    content_type: str,
    content_encoding: str,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    **s3_client_kwargs,
) -> ParquetFile:
    logger.debug(
        f"Reading {s3_url} to PyArrow ParquetFile. "
        f"Content type: {content_type}. Encoding: {content_encoding}"
    )

    if (
        content_type != ContentType.PARQUET.value
        or content_encoding != ContentEncoding.IDENTITY
    ):
        raise ValidationError(
            f"S3 file with content type: {content_type} and "
            f"content encoding: {content_encoding} cannot be read"
            "into pyarrow.parquet.ParquetFile"
        )

    if s3_client_kwargs is None:
        s3_client_kwargs = {}

    kwargs = {}

    if s3_url.startswith("s3://"):
        s3_file_system = create_s3_file_system(s3_client_kwargs)
        kwargs["filesystem"] = s3_file_system

    if pa_read_func_kwargs_provider:
        kwargs = pa_read_func_kwargs_provider(content_type, kwargs)

    logger.debug(f"Pre-sanitize kwargs for {s3_url}: {kwargs}")

    kwargs = sanitize_kwargs_to_callable(ParquetFile.__init__, kwargs)

    logger.debug(
        f"Reading the file from {s3_url} into ParquetFile with kwargs: {kwargs}"
    )
    pqFile, latency = timed_invocation(ParquetFile, s3_url, **kwargs)

    logger.debug(f"Time to get {s3_url} into parquet file: {latency}s")

    return pqFile


def table_size(table: pa.Table) -> int:
    return table.nbytes


def parquet_file_size(table: papq.ParquetFile) -> int:
    return table.metadata.serialized_size


def table_to_file(
    table: pa.Table,
    base_path: str,
    file_system: AbstractFileSystem,
    block_path_provider: BlockWritePathProvider,
    content_type: str = ContentType.PARQUET.value,
    **kwargs,
) -> None:
    """
    Writes the given Pyarrow Table to a file.
    """
    writer = CONTENT_TYPE_TO_PA_WRITE_FUNC.get(content_type)
    if not writer:
        raise NotImplementedError(
            f"Pyarrow writer for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{CONTENT_TYPE_TO_PA_WRITE_FUNC.keys}"
        )
    path = block_path_provider(base_path)
    logger.debug(f"Writing table: {table} with kwargs: {kwargs} to path: {path}")
    writer(table, path, filesystem=file_system, **kwargs)


class RecordBatchTables:
    def __init__(self, batch_size: int):
        """
        Data structure for maintaining a batched list of tables, where each batched table has
        a record count of some multiple of the specified record batch size.

        Remaining records are stored in a separate list of tables.

        Args:
            batch_size: Minimum record count per table to batch by. Batched tables are
             guaranteed to have a record count multiple of the batch_size.
        """
        self._batched_tables: List[pa.Table] = []
        self._batched_record_count: int = 0
        self._remaining_tables: List[pa.Table] = []
        self._remaining_record_count: int = 0
        self._batch_size: int = batch_size

    def append(self, table: pa.Table) -> None:
        """
        Appends a table for batching.

        Table record counts are added to any previous remaining record count.
        If the new remainder record count meets or exceeds the configured batch size record count,
        the remainder will be shifted over to the list of batched tables in FIFO order via table slicing.
        Batched tables will always have a record count of some multiple of the configured batch size.

        Record ordering is preserved from input tables whenever tables are shifted from the remainder
        over to the batched list. Records from Table A will always precede records from Table B,
        if Table A was appended before Table B. Records from the batched list will always precede records
        from the remainders.

        Ex:
            bt = RecordBatchTables(8)
            col1 = pa.array([i for i in range(10)])
            test_table = pa.Table.from_arrays([col1], names=["col1"])
            bt.append(test_table)

            print(bt.batched_records)  # 8
            print(bt.batched)  # [0, 1, 2, 3, 4, 5, 6, 7]
            print(bt.remaining_records)  # 2
            print(bt.remaining)  # [8, 9]

        Args:
            table: Input table to add

        """
        if self._remaining_tables:
            if self._remaining_record_count + len(table) < self._batch_size:
                self._remaining_tables.append(table)
                self._remaining_record_count += len(table)
                return

            records_to_fit = self._batch_size - self._remaining_record_count
            fitted_table = table.slice(length=records_to_fit)
            self._remaining_tables.append(fitted_table)
            self._remaining_record_count += len(fitted_table)
            table = table.slice(offset=records_to_fit)

        record_count = len(table)
        record_multiplier, records_leftover = (
            record_count // self._batch_size,
            record_count % self._batch_size,
        )

        if record_multiplier > 0:
            batched_table = table.slice(length=record_multiplier * self._batch_size)
            # Add to remainder tables to preserve record ordering
            self._remaining_tables.append(batched_table)
            self._remaining_record_count += len(batched_table)

        if self._remaining_tables:
            self._shift_remaining_to_new_batch()

        if records_leftover > 0:
            leftover_table = table.slice(offset=record_multiplier * self._batch_size)
            self._remaining_tables.append(leftover_table)
            self._remaining_record_count += len(leftover_table)

    def _shift_remaining_to_new_batch(self) -> None:
        new_batch = pa.concat_tables(self._remaining_tables)
        self._batched_tables.append(new_batch)
        self._batched_record_count += self._remaining_record_count
        self.clear_remaining()

    @staticmethod
    def from_tables(tables: List[pa.Table], batch_size: int) -> RecordBatchTables:
        """
        Static factory for generating batched tables and remainders given a list of input tables.

        Args:
            tables: A list of input tables with various record counts
            batch_size: Minimum record count per table to batch by. Batched tables are
             guaranteed to have a record count multiple of the batch_size.

        Returns: A batched tables object

        """
        rbt = RecordBatchTables(batch_size)
        for table in tables:
            rbt.append(table)
        return rbt

    @property
    def batched(self) -> List[pa.Table]:
        """
        List of tables batched and ready for processing.
        Each table has N records, where N records are some multiple of the configured records batch size.

        For example, if the configured batch size is 5, then a list of batched tables
        could have the following record counts: [60, 5, 30, 10]

        Returns: a list of batched tables

        """
        return self._batched_tables

    @property
    def batched_record_count(self) -> int:
        """
        The number of total records from the batched list.

        Returns: batched record count

        """
        return self._batched_record_count

    @property
    def remaining(self) -> List[pa.Table]:
        """
        List of tables carried over from table slicing during the batching operation.
        The sum of all record counts in the remaining tables is guaranteed to be less than the configured batch size.

        Returns: a list of remaining tables

        """
        return self._remaining_tables

    @property
    def remaining_record_count(self) -> int:
        """
        The number of total records from the remaining tables list.

        Returns: remaining record count

        """
        return self._remaining_record_count

    @property
    def batch_size(self) -> int:
        """
        The configured batch size.

        Returns: batch size

        """
        return self._batch_size

    def has_batches(self) -> bool:
        """
        Checks if there are any currently batched tables ready for processing.

        Returns: true if batched records exist, otherwise false

        """
        return self._batched_record_count > 0

    def has_remaining(self) -> bool:
        """
        Checks if any remaining tables exist after batching.

        Returns: true if remaining records exist, otherwise false

        """
        return self._remaining_record_count > 0

    def evict(self) -> List[pa.Table]:
        """
        Evicts all batched tables from this object and returns them.

        Returns: a list of batched tables

        """
        evicted_tables = [*self.batched]
        self.clear_batches()
        return evicted_tables

    def clear_batches(self) -> None:
        """
        Removes all batched tables and resets batched records.

        """
        self._batched_tables.clear()
        self._batched_record_count = 0

    def clear_remaining(self) -> None:
        """
        Removes all remaining tables and resets remaining records.

        """
        self._remaining_tables.clear()
        self._remaining_record_count = 0
